from __future__ import print_function
from __future__ import absolute_import
from builtins import range
import numpy as np
import pandas as pd
import multiprocessing as mp
from munch import Munch
import os, random
from six import string_types
from tensorflow import keras

from .image_utils import *
from .generic import *
from .tensor_ops import *

# GENERATORS

class DataGeneratorDisk(keras.utils.Sequence):
    """
    Generates data for training Keras models
    - inherits from keras.utils.Sequence
    - reads images from disk and applies `process_fn` to each
    - `process_fn` needs to ensure that processed images are of the same size
    - on __getitem__() returns an ND-array containing `batch_size` images

    ARGUMENTS
    * ids (pandas.dataframe): table containing image names, and output variables
    * data_path  (string):    path of image folder
    * batch_size (int):       how many images to read at a time
    * shuffle (bool):         randomized reading order
    * ids_fn (function):      returns an updated `ids` table, replacing `self.ids` at the end of each epoch
    * process_fn (function):  function applied to each image as it is read
    * read_fn (function):     function used to read data from a file (returns numpy.array)
                              if None, image_utils.read_image() is used (default)
    * deterministic (None, int):   random seed for shuffling order
    * inputs (list of strings):    column names from `ids` containing image names
    * inputs_df (list of strings): column names from `ids`, returns values from the DataFrame itself
    * outputs (list of strings):   column names from `ids`
    * verbose (bool):             logging verbosity
    * fixed_batches (bool):       only return full batches, ignore the last incomplete batch if needed
    * process_args (dictionary):  dict of corresponding `ids` columns for `inputs`
                                  containing arguments to pass to `process_fn`
    * group_names (strings list): read only from specified sub-paths (groups), or from any if `group_names` is None
                                   `group_names` are randomly sampled from meta-groups
                                   i.e. when group_names = [[group_names_1], [group_names_2]]
    * random_group (bool): read inputs from a random group for every image

    """
    def __init__(self, ids, data_path, **args):
        params_defa = Munch(ids           = ids,   data_path= data_path,
                            batch_size    = 32,    shuffle  = True,
                            input_shape   = None,  ids_fn   = None,
                            process_fn    = None,  read_fn  = None,
                            deterministic = None,  inputs   = [],
                            inputs_df     = None,  outputs  = [],
                            verbose       = False, fixed_batches = False,
                            random_group  = False, group_names   = None,
                            process_args  = {},    group_by = None)
        check_keys_exist(args, params_defa)
        params = updated_dict(params_defa, **args)  # update only existing
        params.deterministic = {True: 42, False: None}.\
                               get(params.deterministic,
                                   params.deterministic)
        params.process_args = params.process_args or {}
        params.group_names  = params.group_names or ['']
        self.__dict__.update(**params)  # set all as self.<param>
        if self.verbose>1:
            print('Initialized DataGeneratorDisk')
        self.on_epoch_end()  # initialize indexes

    def __len__(self):
        """Get the number of batches per epoch"""
        if not self.group_by:
            round_op = np.floor if self.fixed_batches else np.ceil
            return int(round_op(len(self.ids)*1. / self.batch_size))
        else:
            return int(self.ids_index.batch_index.max()+1)

    def __getitem__(self, index):
        """Generate a batch of data"""
        if self.verbose:
            show_progress(index, len(self), prefix='Generating batches')

        ids_batch = self.ids[self.ids_index.batch_index==index]

        # reshuffle to remove ordering by index
        if self.shuffle:
            ids_batch = ids_batch.reset_index(drop=True).\
                        sample(frac=1, random_state=self.deterministic)

        return self._data_generation(ids_batch)

    def on_epoch_end(self):
        """Updates batch selection after each epoch"""

        if self.ids_fn is not None:
            self.ids = self.ids_fn()

        if self.group_by:
            group_dict = dict(group_by=self.ids[self.group_by])
        else:
            group_dict = None

        self.ids_index = pd.DataFrame(group_dict,
                                      index=self.ids.index.copy())
        self.ids_index['batch_index'] = -1

        # initialize batch indexes
        index = 0
        selectable = self.ids_index.batch_index == -1
        while selectable.sum():
            ids_sel = self.ids_index[selectable]
            if self.group_by:
                group_by_value = ids_sel.group_by.sample(1,
                        random_state=self.deterministic).values[0]
                ids_sel = ids_sel[ids_sel.group_by==group_by_value]

            batch_size_max = min(self.batch_size, len(ids_sel))
            if self.shuffle:
                ids_batch = ids_sel.sample(batch_size_max,
                                           random_state=self.deterministic)
            else:
                ids_batch = ids_sel.iloc[:batch_size_max]

            self.ids_index.loc[ids_batch.index, 'batch_index'] = index
            index += 1
            selectable = self.ids_index.batch_index == -1

    def _access_field(self, ids_batch, accessor):
        if isinstance(ids_batch[accessor].values[0][0], np.ndarray):
            return np.stack(ids_batch[accessor].values.squeeze(), axis=0)
        else:
            return ids_batch[accessor].values

    def _read_data(self, ids_batch, accessor):
        X = []
        if accessor:
            assert isinstance(accessor, list) or callable(accessor),\
            'Generator inputs/outputs must be of type list, or function'

            if callable(accessor):
                X = accessor(ids_batch)
            elif isinstance(accessor, list):
                if all(isinstance(a, list) for a in accessor):
                    X = [self._access_field(ids_batch,a) for a in accessor]
                else:
                    assert all(not isinstance(a, list) for a in accessor)
                    X = [self._access_field(ids_batch,accessor)]
            else:
                 raise Exception('Wrong generator input/output specifications')

        return X

    def _data_generation(self, ids_batch):
        """Generates image-stack + outputs containing batch_size samples"""
        params = self
        np.random.seed(params.deterministic)

        y = self._read_data(ids_batch, params.outputs)
        X_list = self._read_data(ids_batch, params.inputs_df)

        assert isinstance(params.inputs, list),\
        'Generator inputs/outputs must be of type list'

        # group_names are randomly sampled from meta-groups
        # i.e. when group_names = [[group_names1], [group_names2]]
        group_names = params.group_names
        if isinstance(group_names[0], list):
            idx = np.random.randint(0, len(group_names))
            group_names = group_names[idx]

        if isinstance(params.data_path, string_types):
            # get data for each input and add it to X_list
            for group_name in group_names:
                if params.random_group:
                    group_path = os.path.join(params.data_path, group_name) or '.'
                    subdir_names = [f for f in os.listdir(group_path)
                                    if os.path.isdir(os.path.join(group_path, f))]
                    subgroup_name = random.choice(subdir_names)
                else:
                    subgroup_name = ''

                for input_name in params.inputs:
                    data = []
                    # read the data from disk into a list
                    for row in ids_batch.itertuples():
                        input_data = os.path.join(group_name, subgroup_name, getattr(row, input_name))
                        if params.read_fn is None:
                            file_path = os.path.join(params.data_path, input_data)
                            file_data = read_image(file_path)
                        else:
                            file_data = params.read_fn(input_data, params)
                        data.append(file_data)

                    # column name for the arguments to `process_fn`
                    args_name = params.process_args.get(input_name, None)

                    # if needed, process each image, and add to X_list (inputs list)
                    if params.process_fn not in [None, False]:
                        data_list = []
                        for i, row in enumerate(ids_batch.itertuples()):
                            args = []
                            if args_name is not None:
                                args = [getattr(row, name) for name in force_list(args_name)]
                            data_i = params.process_fn(data[i], *args)
                            data_list.append(force_list(data_i))

                        # transpose list, sublists become batches
                        data_list = zip(*data_list)

                        # for each sublist of arrays
                        data_arrays = []
                        for batch_list in data_list:
                            batch_arr = np.float32(np.stack(batch_list))
                            data_arrays.append(batch_arr)

                        X_list.extend(data_arrays)
                    else:
                        data_array = np.float32(np.stack(data))
                        X_list.append(data_array)

        np.random.seed(None)
        return (X_list, y)


class DataGeneratorHDF5(DataGeneratorDisk):
    """
    Generates data for training Keras models
    - similar to the `DataGeneratorDisk`, but reads data instances from an HDF5 file e.g. images, features
    - inherits from `DataGeneratorDisk`, a child of keras.utils.Sequence
    - applies `process_fn` to each data instance
    - `process_fn` needs to ensure a fixed size for processed data instances
    - on __getitem__() returns an ND-array containing `batch_size` data instances

    ARGUMENTS
    * ids (pandas.dataframe):    table containing data instance names, and output variables
    * data_path  (string):       path of HDF5 file
    * batch_size (int):          how many instances to read at a time
    * shuffle (bool):            randomized reading order
    * ids_fn (function):         returns an updated `ids` table, replacing `self.ids` at the end of each epoch
    * process_fn (function):     function applied to each data instance as it is read
    * deterministic (None, int): random seed for shuffling order
    * inputs (strings list):     column names from `ids` containing data instance names, read from `data_path`
    * inputs_df (strings list):  column names from `ids`, returns values from the DataFrame itself
    * outputs (strings list):    column names from `ids`, returns values from the DataFrame itself
    * verbose (bool):            logging verbosity
    * fixed_batches (bool):      only return full batches, ignore the last incomplete batch if needed
    * process_args (dictionary): dict of corresponding `ids` columns for `inputs`
                                 containing arguments to pass to `process_fn`
    * group_names (strings list): read only from specified groups, or from any if `group_names` is None
                                   `group_names` are randomly sampled from meta-groups
                                   i.e. when group_names = [[group_names_1], [group_names_2]]
    * random_group (bool): read inputs from a random group for every data instance
    """
    def __init__(self, ids, data_path, **args):
        params_defa = Munch(ids         = ids,   data_path     = data_path, deterministic = False,
                            batch_size  = 32,    shuffle       = True,   inputs        = [],
                            inputs_df   = None,  outputs       = [],     memory_mapped = False,
                            verbose     = False, fixed_batches = False,  random_group  = False,
                            process_fn  = None,  process_args  = None,   group_names   = None,
                            input_shape = None,  group_by      = None,   ids_fn        = None)

        check_keys_exist(args, params_defa)
        params = updated_dict(params_defa, **args) # update only existing
        params.process_args  = params.process_args or {}
        params.group_names   = params.group_names or [None]
        params.deterministic = {True: 42, False: None}.\
                                get(params.deterministic,
                                    params.deterministic)
        self.__dict__.update(**params)  # set all as self.<param>
        if self.verbose>1:
            print('Initialized DataGeneratorHDF5')
        self.on_epoch_end()  # initialize indexes

    def _data_generation(self, ids_batch):
        """Generates data containing batch_size samples"""
        params = self
        np.random.seed(params.deterministic)

        y = self._read_data(ids_batch, params.outputs)
        X_list = self._read_data(ids_batch, params.inputs_df)

        assert isinstance(params.inputs, list),\
        'Generator inputs/outputs must be of type list'

        if isinstance(params.data_path, string_types):
            with H5Helper(params.data_path, file_mode='r',
                          memory_mapped=params.memory_mapped) as h:
                # group_names are randomly sampled from meta-groups
                # i.e. when group_names = [[group_names1], [group_names2]]
                group_names = params.group_names
                if isinstance(group_names[0], list):
                    idx = np.random.randint(0, len(group_names))
                    group_names = group_names[idx]

                # get data for each input and add it to X_list
                for group_name in group_names:
                    for input_name in params.inputs:
                        # get data
                        names = list(ids_batch.loc[:,input_name])
                        if params.random_group:
                            data = h.read_data_random_group(names)
                        elif group_name is None:
                            data = h.read_data(names)
                        else:
                            data = h.read_data(names, group_names=[group_name])[0]
                        if data.dtype != np.float32:
                            data = data.astype(np.float32)

                        # column name for the arguments to `process_fn`
                        args_name = params.process_args.get(input_name, None)

                        # add to X_list
                        if params.process_fn not in [None, False]:
                            data_new = None
                            for i, row in enumerate(ids_batch.itertuples()):
                                arg = [] if args_name is None else [getattr(row,args_name)]
                                data_i = params.process_fn(data[i,...], *arg)
                                if data_new is None:
                                    data_new = np.zeros((len(data),)+data_i.shape,
                                                       dtype=np.float32)
                                data_new[i,...] = data_i
                            X_list.append(data_new)
                        else:
                            X_list.append(data)

        np.random.seed(None)
        return (X_list, y)


class GeneratorStack(keras.utils.Sequence):
    """
    Creates an aggregator generator that feeds from multiple generators.
    """
    def __init__(self, generator_list):
        self.gens = generator_list

    def __len__(self):
        """Number of batches per epoch"""
        return len(self.gens[0])

    def __getitem__(self, index):
        """Generate one batch of data"""
        X, y = [],[]
        for g in self.gens:
            X_, y_ = g[index]
            X.extend(X_)
            y.extend(y_)
        return (X, y)




from __future__ import print_function
from __future__ import absolute_import
from builtins import map
from builtins import str
from builtins import zip
from builtins import range
from builtins import object
import os, sys, numbers, glob, shutil, inspect, copy
import multiprocessing as mp, pandas as pd, numpy as np
from pprint import pprint
from munch import Munch
from collections import OrderedDict

from tensorflow import keras
from tensorflow.keras.callbacks import TensorBoard, ModelCheckpoint, Callback, EarlyStopping
from tensorflow.keras import optimizers
from tensorflow.keras.models import Model, load_model
from tensorflow.keras.utils import multi_gpu_model

from .generators import *
from .generic import *


class ModelHelper(object):
    """
    Wrapper class that simplifies default usage of Keras for training, testing, logging,
    storage of models by removing the need for some repetitive code segments (boilerplate).

    Encapsulates a Keras model and its configuration, generators for feeding the model
    during training, validation and testing, logging to TensorBoard, saving/loading
    model configurations to disk, and activations from within a model.

    When operating on a model, generators can be instantiated by the ModelHelper or pre-defined
    by the user and passed to the train/predict methods. Generators, rely on DataFrame objects
    (usually named `ids`, as they contain the IDs of data rows) for extracting data instances
    for all operations (train/validation/test).
    """
    def __init__(self, model, root_name, ids,
                 gen_params={}, verbose=False, **params):
        """
        * model:      Keras model, compilation will be done at runtime.
        * root_name:  base name of the model, extended with
                      configuration parameters when saved
        * ids:        DataFrame table, used by generators
        * gen_params: dict for parameters for the generator
                      defaults (shuffle = True, process_fn = False,
                                deterministic = False, verbose = False)
        * verbose:    verbosity

        OTHER PARAMS
        lr   = 1e-4,               # learning rate
        loss = "MSE",              # loss function
        loss_weights   = None,     # loss weights, if multiple losses are used
        metrics        = [],       #
        class_weights  = None,     # class weights for unbalanced classes
        multiproc      = True,     # multi-processing params
        workers        = 5,        #
        max_queue_size = 10,       #
        monitor_metric      = 'val_loss',  # monitoring params
        monitor_mode        = 'min',       #
        early_stop_patience = 20,          #
        checkpoint_period   = 1,           #
        save_best_only      = True,        #
        optimizer      = optimizers.Adam(),  # optimizer object
        write_graph    = False,              # TensorBoard params
        write_images   = False,              #
        logs_root      = None,               # TensorBoard logs
        models_root    = 'models/',          # saved models path
        features_root  = 'features/',        # saved features path (by `save_activations`)
        gen_class      = None                # generator class
                                             # inferred from self.gen_params.data_path
        callbacks      = None                # option to customize callbacks list
        """
        self.model = model
        self.ids = ids
        self.verbose = verbose
        self.model_name = ShortNameBuilder(prefix=root_name+'/',
                                           sep=(':', ' '))
        self.model_cpu = None

        self.gen_params = Munch(shuffle       = True,  process_fn = False,
                                deterministic = False, verbose    = verbose,
                                batch_size    = 32)
        self.params = Munch(lr   = 1e-4,               # learning rate
                            loss = "MSE",              # loss function
                            loss_weights   = None,     # loss weights, if multiple losses are used
                            metrics        = [],       #
                            class_weights  = None,     # class weights for unbalanced classes

                            multiproc      = True,     # multi-processing params
                            workers        = 2,        #
                            max_queue_size = 10,       #

                            monitor_metric      = 'val_loss',  # monitoring params
                            monitor_mode        = 'min',       #
                            early_stop_patience = 20,          #
                            checkpoint_period   = 1,           #
                            save_best_only      = True,        #

                            optimizer      = optimizers.Adam(),  # optimizer object, its parameters
                                                                 # can changed during runtime
                            write_graph    = False,              # TensorBoard params
                            write_images   = False,              #
                            logs_root      = None,               # TensorBoard logs
                            models_root    = 'models/',          # saved models path
                            features_root  = 'features/',        # saved features path (by `save_activations`)
                            gen_class      = None,               # generator class inferred from self.gen_params.data_path
                            callbacks      = None                # option to customize callbacks list
                            )

        for key in list(params.keys()):
            if key not in list(self.params.keys()):
                raise Exception('Undefined parameter:' + key)

        self.gen_params.update(gen_params)
        self.params = updated_dict(self.params, **params)

        # infer default generator class to use
        # if params is not set yet
        if self.params.gen_class is None:
            # if has 'data_path' attribute
            if getattr(self.gen_params, 'data_path', None) is not None:
                if self.gen_params.data_path[-3:] == '.h5':
                    self.params.gen_class = DataGeneratorHDF5
                else:
                    self.params.gen_class = DataGeneratorDisk
            else:
                raise ValueError("Cannot infer generator class")

        self.callbacks = None
        self.set_model_name()

    def update_name(self, **kwargs):
        """Propagate configuration: update model_name and callbacks"""
        # update configuration parameters
        self.set_model_name()
        # update new parameters
        self.model_name(**kwargs)

        # always use custom callbacks, if defined
        if isinstance(self.params.callbacks, list):
            self.callbacks = self.params.callbacks
        else: # reset callbacks to use new model name
            self.callbacks = self.set_callbacks()

        return self.model_name

    def set_model_name(self):
        """Update model name based on parameters in self. gen_params, params and model"""
        h = self
        tostr = lambda x: str(x) if x is not None else '?'
        format_size = lambda x: '[{}]'.format(','.join(map(tostr, x)))
        loss2str = lambda x: (x if isinstance(x, (str, bytes)) else x.__name__)[:9]

        loss = h.params.loss
        if not isinstance(loss, dict):
            loss_str = loss2str(loss)
        else:
            loss_str = '[%s]' % ','.join(map(loss2str, list(loss.values())))
        i = '{}{}'.format(len(h.model.inputs),
                           format_size(h.gen_params.input_shape))
        if h.model.outputs:
            o = '{}{}'.format(len(h.model.outputs),
                              format_size(h.model.outputs[0].shape[1:].as_list()))
        else: o = ''
        name = dict(i   = i,
                    o   = o,
                    l   = loss_str,
                    bsz = h.gen_params.batch_size)
        self.model_name.update(name)

        return self.model_name

    def set_callbacks(self):
        """Setup callbacks"""

        p = self.params
        if p.logs_root is not None:
            log_dir = os.path.join(p.logs_root, self.model_name())
            tb_callback = TensorBoard(log_dir        = log_dir,
                                      write_graph    = p.write_graph,
                                      histogram_freq = 0,
                                      write_images   = p.write_images)
            tb_callback.set_model(self.model)
        else:
            tb_callback = None

#        separator = self.model_name._ShortNameBuilder__sep[1]
        best_model_path = os.path.join(p.models_root,
                                       self.model_name() + '_best_weights.h5')
        make_dirs(best_model_path)
        checkpointer = ModelCheckpoint(filepath = best_model_path, verbose=0,
                                       monitor  = p.monitor_metric,
                                       mode     = p.monitor_mode,
                                       period   = p.checkpoint_period,
                                       save_best_only    = p.save_best_only,
                                       save_weights_only = True)

        earlystop = EarlyStopping(monitor  = p.monitor_metric,
                                  patience = p.early_stop_patience,
                                  mode     = p.monitor_mode)

        return [earlystop, checkpointer] + ([tb_callback] if tb_callback else [])


    def _updated_gen_params(self, **kwargs):
        params = copy.deepcopy(self.gen_params)
        params.update(kwargs)
        return params

    def make_generator(self, ids, subset=None, **kwargs):
        """
        Create a generator of `self.params.gen_class` using new `ids`.

        * ids:    DataFrame table
        * subset: select a subset of rows based on the `set` columns
        * kwargs: updated parameters of the generator
        :return: Sequence (generator)
        """
        params = self._updated_gen_params(**kwargs)
        if subset:
            ids_ = ids[ids.set==subset]
        else:
            ids_ = ids
        return self.params.gen_class(ids_, **params)

    def test_generator(self, input_idx=0):
        """
        Basic utility to run the generator for one or more data instances in `ids`.
        Useful for testing the default generator functionality.

        * input_idx: scalar or 2-tuple of indices
        :return: generated_batch
        """
        if not isinstance(input_idx, (list, tuple)):
            ids_gen = self.ids.iloc[input_idx:input_idx+1]
        else:
            ids_gen = self.ids.iloc[input_idx[0]:input_idx[1]]

        gen = self.make_generator(ids_gen)
        x = gen[0]  # generated data
        print('First batch sizes:'); print_sizes(x)
        return x

    def set_multi_gpu(self, gpus=None):
        """
        Enable multi-GPU processing.
        Creates a copy of the CPU model and calls `multi_gpu_model`.

        * gpus: number of GPUs to use, defaults to all.
        """
        self.model_cpu = self.model
        self.model = multi_gpu_model(self.model, gpus=gpus)

    def compile(self):
        self.model.compile(optimizer    = self.params.optimizer,
                           loss         = self.params.loss,
                           loss_weights = self.params.loss_weights,
                           metrics      = self.params.metrics)

    def train(self, train_gen=True, valid_gen=True,
              lr=1e-4, epochs=1, valid_in_memory=False,
              recompile=True, verbose=True):
        """
        Run training iterations on existing model.
        Initializes `train_gen` and `valid_gen` if not defined.

        * train_gen: train generator
        * valid_gen: validation generator
        * lr:        learning rate
        * epochs:    number of epochs
        :return:     training history from self.model.fit_generator()
        """
        ids = self.ids
        print('Training model:', self.model_name())

        if train_gen is True:
            train_gen = self.make_generator(ids[ids.set == 'training'])
        if valid_gen is True:
            valid_gen = self.make_generator(ids[ids.set == 'validation'],
                                            shuffle       = True,
                                            deterministic = True)

        if recompile:
            if lr: self.params.lr = lr
            self.params.optimizer =\
                update_config(self.params.optimizer,
                              lr=self.params.lr)
            self.compile()

        if self.verbose:
            print('\nGenerator parameters:')
            print('---------------------')
            pretty(self.gen_params)
            print('\nMain parameters:')
            print('----------------')
            pretty(self.params)
            print('\nLearning')

        if valid_in_memory:
            valid_gen.batch_size = len(valid_gen.ids)
            valid_gen.on_epoch_end()
            valid_data = valid_gen[0]
            valid_steps = 1
        else:
            valid_data = valid_gen
            if issubclass(type(valid_gen),
                          keras.utils.Sequence):
                valid_steps = len(valid_gen)
            else:
                valid_steps = 1

        # always use custom callbacks, if defined
        if isinstance(self.params.callbacks, list):
            self.callbacks = self.params.callbacks
        # set callbacks if first run, otherwise do not reset
        elif self.callbacks is None:
            self.callbacks = self.set_callbacks()

        if issubclass(type(train_gen),
                      keras.utils.Sequence):
            # train using the generator
            history = self.model.fit_generator(
                             train_gen, epochs = epochs,
                             steps_per_epoch   = len(train_gen),
                             validation_data   = valid_data,
                             validation_steps  = valid_steps,
                             workers           = self.params.workers,
                             callbacks         = self.callbacks,
                             max_queue_size    = self.params.max_queue_size,
                             class_weight      = self.params.class_weights,
                             use_multiprocessing = self.params.multiproc,
                             verbose             = verbose)
        else:
            # training data is passed in train_gen
            X, y = train_gen
            steps_per_epoch = X.shape[0]//self.gen_params.batch_size

            history = self.model.fit(X, y,
                             batch_size      = self.gen_params.batch_size,
                             epochs          = epochs,
                             callbacks       = self.callbacks,
                             validation_data = valid_data,
                             shuffle         = self.gen_params.shuffle,
                             class_weight    = self.params.class_weights,
                             verbose         = verbose)

        return history

    def clean_outputs(self, force=False):
        """
        Delete training logs or models created by the current helper configuration.
        Identifies the logs by the configuration paths and `self.model_name`.
        Asks for user confirmation before deleting any files.
        """
        log_dir = os.path.join(self.params.logs_root, self.model_name())
        model_path = os.path.join(self.params.models_root, self.model_name()) + '*.h5'
        model_path = model_path.replace('[', '[[]')
        model_files = glob.glob(model_path)

        if os.path.exists(log_dir):
            print('Found logs:')
            print(log_dir)
            if force or raw_confirm('Delete?'):
                print('Deleting', log_dir)
                shutil.rmtree(log_dir)
        else:
            print('(No logs found)')

        if model_files:
            print('Found model(s):')
            print(model_files)
            if force or raw_confirm('Delete?'):
                for mf in model_files:
                    print('Deleting', mf)
                    os.unlink(mf)
        else:
            print('(No models found)')

    def predict(self, test_gen=None, output_layer=None,
                repeats=1, batch_size=None, remodel=True):
        """
        Predict on `test_gen`.

        * test_gen:     generator used for prediction
        * output_layer: layer at which activations are computed (defaults to output)
        * repeats:      how many times the prediction is repeated (with different augmentation)
        * batch_size:   size of each batch
        * remodel:      if `true` then change model such that new output is `output_layer`
        :return:        if `repeats` == 1 then `np.ndarray` else `list[np.ndarray]`
        """
        if not test_gen:
            params_test = self._updated_gen_params(shuffle       = False,
                                                   fixed_batches = False)
            if batch_size: params_test.batch_size = batch_size
            test_gen = self.params.gen_class(self.ids[self.ids.set == 'test'],
                                             **params_test)
        if output_layer is not None and remodel:
                # get last partial-matching layer
                layer_name = [l.name for l in self.model.layers
                              if output_layer in l.name][-1]
                output_layer = self.model.get_layer(layer_name)
                print('Output of layer:', output_layer.name)
                if isinstance(output_layer, Model):
                    outputs = output_layer.outputs[0]
                else:
                    outputs = output_layer.output
                print('Output tensor:', outputs)
                model = Model(inputs  = self.model.input,
                              outputs = outputs)
        else:
            model = self.model

        preds = []
        for i in range(repeats):
            y_pred = model.predict_generator(test_gen, workers=1, verbose=0,
                                             use_multiprocessing=False)
            if not remodel and output_layer is not None:
                y_pred = dict(list(zip(model.output_names, y_pred)))[output_layer]
            preds.append(y_pred)
        return preds[0] if repeats == 1 else preds

    def validate(self, valid_gen=True, verbose=2,
                 batch_size=32, recompile=True):
        if valid_gen is True:
            ids = self.ids
            valid_gen = self.make_generator(ids[ids.set == 'validation'],
                                            shuffle       = True,
                                            deterministic = True)
        print('Validating performance')
        if recompile: self.compile()

        if issubclass(type(valid_gen), keras.utils.Sequence):
            r = self.model.evaluate_generator(valid_gen,
                                              verbose=verbose)
        else:
            X_valid, y_valid = valid_gen
            r = self.model.evaluate(X_valid, y_valid,
                                    batch_size=batch_size,
                                    verbose=verbose)
        perf_metrics = dict(list(zip(self.model.metrics_names,
                                     force_list(r))))
        if verbose==2:
            pretty(OrderedDict((k, perf_metrics[k])
                               for k in sorted(perf_metrics.keys())))
        return perf_metrics

    def set_trainable(self, index):
        """
        Convenience method to set trainable layers.
        Layers up to `index` are frozen; the remaining
        after `index` are set as trainable.
        """
        for layer in self.model.layers[:index]:
            layer.trainable = False
        for layer in self.model.layers[index:]:
            layer.trainable = True

    def load_model(self, model_name='', best=True,
                   from_weights=True, by_name=False, verbose=1):
        """
        Load model from file.

        * model_name:   new model name, otherwise self.model_name()
        * best:         load the best model, or otherwise final model
        * from_weights: from weights, or from full saved model
        * by_name:      load layers by name
        :return:        true if model was loaded successfully, otherwise false
        """
        model_name = model_name or self.model_name()
        model_file_name = (model_name + ('_best' if best else '_final') +
                          ('_weights' if from_weights else '') + '.h5')
        model_path = os.path.join(self.params.models_root, model_file_name)
        if not os.path.exists(model_path):
            if verbose:
                print('Model NOT loaded:', model_file_name, 'does not exist')
            return False
        else:
            if from_weights:
                self.model.load_weights(model_path, by_name=by_name)
                if verbose:
                    print('Model weights loaded:', model_file_name)
            else:
                self.model = load_model(model_path)
                if verbose:
                    print('Model loaded:', model_file_name)
            return True

    def save_model(self, weights_only=False, model=None,
                   name_extras='', best=False, verbose=1):
        """
        Save model to HDF5 file.

        * weights_only: save only weights,
                        or full model otherwise
        * model:        specify a particular model instance,
                        otherwise self.model is saved
        * name_extras:  append this to model_name
        * best:         save as best model, otherwise final model
        """
        model = model or self.model
        if verbose:
            print('Saving model', model.name, 'spanning',\
                  len(self.model.layers), 'layers')
        model_type = 'best' if best else 'final'
        if weights_only:
            model_file = self.model_name() + name_extras + '_' + model_type + '_weights.h5'
            model.save_weights(os.path.join(self.params.models_root, model_file))
            if verbose:
                print('Model weights saved:', model_file)
        else:
            model_file = self.model_name() + name_extras + '_' + model_type + '.h5'
            model.compile(optimizer=self.params.optimizer, loss="mean_absolute_error")
            model.save(os.path.join(self.params.models_root, model_file))
            if verbose:
                print('Model saved:', model_file)

    def save_activations(self, output_layer=None, file_path=None, ids=None,
                         groups=1, verbose=False, overwrite=False, name_suffix='',
                         save_as_type=np.float32, postprocess_fn=None):
        """
        Save activations from a particular `output_layer` to an HDF5 file.

        * output_layer:   if not None, layer name, otherwise use the model output
        * file_path:      HDF5 file path
        * ids:            data entries to compute the activations for, defaults to `self.ids`
        * groups:         a number denoting the number of augmentation repetitions, or list of group names
        * verbose:        verbosity
        * overwrite:      overwrite HDF5 file
        * name_suffix:    append suffix to name of file (before ext)
        * save_as_type:   save as a different data type, defaults to np.float32
        * postprocess_fn: post-processing function to apply to the activations
        """
        if ids is None:
            ids = self.ids
        if isinstance(groups, numbers.Number):
            groups_count = groups
            groups_list = list(map(str, list(range(groups))))
        else:
            groups_count = len(groups)
            groups_list  = list(map(str, groups))

        if file_path is None:
            short_name = self.model_name.subset(['i', 'o'])
            short_name(grp = groups_count,
                       lay = (output_layer or 'final'))
            file_path = os.path.join(self.params.features_root,
                                     str(short_name) + name_suffix + '.h5')
            make_dirs(file_path)

        params = self._updated_gen_params(shuffle       = False,
                                          verbose       = verbose,
                                          fixed_batches = False,
                                          group_by      = None)
        if verbose>1:
            print('Saving activations for layer:', (output_layer or 'final'))
            print('file:', file_path)

        data_gen = self.make_generator(ids, **params)

        for group_name in groups_list:
            activ = self.predict(data_gen,
                                 output_layer = output_layer)
            activ = activ.astype(save_as_type)
            if len(activ.shape)==1:
                activ = np.expand_dims(activ, 0)
            if postprocess_fn:
                activ = postprocess_fn(activ)

            with H5Helper(file_path,
                          overwrite = overwrite,
                          verbose   = verbose) as h:
                if groups == 1:
                    h.write_data(activ, list(ids.loc[:,data_gen.inputs[0]]))
                else:
                    h.write_data([activ], list(ids.loc[:,data_gen.inputs[0]]),
                                 group_names=[group_name])
            del activ

    def save_features(self, process_gen, batch_size = 1024,
                      save_as_type = np.float32, overwrite = False):
        """
        Save augmented features to HDF5 file.

        * process_gen: if function: applies `process_gen` as `self.gen_params.preprocess_fn`
                       if generator: defines `preprocess_fn` functions to use for each augmentation
        * batch_size: batch size used for storing activations (in an `np.ndarray`)
        * save_as_type: convert the activations to this type, to reduce required storage
        * overwrite: overwrite features file
        """
        ids = self.ids
        print('[Saving features]')

        if not inspect.isgeneratorfunction(process_gen):
            process_gen = [(process_gen,None)]
        else:
            process_gen = process_gen()

        first_verbose = 2
        for (process_fn, args) in process_gen:
            self.gen_params.process_fn = process_fn
            if args:
                if isinstance(args, dict):
                    arg_str = ', '.join(['{}:{}'.format(*a) for a in args.items()])
                else:
                    arg_str = str(args)
                print('Augmentation args "' + arg_str + '"')

            numel = len(ids)
            for i in range(0,numel,batch_size):
                istop = min(i+batch_size, numel)
                print('\nImages',i,':',istop)
                ids_batch = ids.iloc[i:istop]

                self.save_activations(ids=ids_batch, verbose = first_verbose,
                                      groups       = [arg_str] if args else 1,
                                      save_as_type = save_as_type,
                                      overwrite    = overwrite)
                overwrite = False
                first_verbose = 1

class TensorBoardWrapper(TensorBoard):
    """Sets the self.validation_data property for use with TensorBoard callback."""

    def __init__(self, valid_gen, **kwargs):
        super(TensorBoardWrapper, self).__init__(**kwargs)
        self.valid_gen = valid_gen  # The validation generator.

    def on_epoch_end(self, epoch, logs):
        # Fill in the `validation_data` property.
        X, y = self.valid_gen[0]
        X = np.float32(X);
        y = np.float32(y)
        sample_weights = np.ones(X.shape[0], dtype=np.float32)
        self.validation_data = [X, y, sample_weights, np.float32(0.0)]
        return super(TensorBoardWrapper, self).on_epoch_end(epoch, logs)


# MISC helper functions

def get_layer_index(model, name):
    """Get index of layer by name"""
    for idx, layer in enumerate(model.layers):
        if layer.name == name:
            return idx

def get_activations(im, model, layer):
    """Get activations from `layer` in `model` using `K.function`"""
    if len(im.shape) < 4:
        im = np.expand_dims(im, 0)
    inputs = [K.learning_phase()] + model.inputs
    fn = K.function(inputs, [layer.output])
    act = fn([0] + [im])
    act = np.squeeze(act)
    return act




from __future__ import print_function
from __future__ import division
from builtins import zip
from builtins import str
from builtins import range
from past.utils import old_div
from builtins import object
import numpy as np, pandas as pd
import multiprocessing as mp
import os, scipy, h5py, time, sys
import urllib, shutil, subprocess
from munch import Munch
from sklearn.model_selection import train_test_split

if sys.version_info.major == 2:
    input = raw_input

# Helps with the DataGeneratorHDF5
class H5Helper(object):
    """
    Read/Write named data sets from/to an HDF5 file.
    The structure of the data inside the HDF5 file is:
    'group_name/dataset_name' or 'dataset_name'

    Enables reading from random groups e.g. 'augmentation_type/file_name'
    such that the helper can be used during training Keras models.
    """
    def __init__(self, file_name, file_mode=None,
                 memory_mapped=False, overwrite=False,
                 backing_store=False, verbose=False):
        """
        * file_name: HDF5 file path
        * file_mode: one of 'a','w','r','w+','r+'
        * memory_mapped: enables memory mapped backing store
        * overwrite: over-write existing file
        * backing_store: use another backing store
        * verbose: verbosity
        """
        self.hf = None
        self.file_name = file_name
        self.verbose = verbose
        self.memory_mapped = memory_mapped
        self.backing_store = backing_store
        self._lock = mp.Lock()
        _file_mode = file_mode or ('w' if overwrite else 'a')
        with self._lock:
            if memory_mapped:
                # memory mapping via h5py built-ins
                self.hf = h5py.File(file_name, _file_mode, driver='core',
                                    backing_store = backing_store)
            else:
                self.hf = h5py.File(file_name, _file_mode)

    def _write_datasets(self, writer, data, dataset_names):
        # internal to the class
        for i in range(len(data)):
            if self.verbose:
                show_progress(i, len(data), prefix='Writing datasets')
            writer.create_dataset(dataset_names[i], data=data[i, ...])

    def write_data(self, data, dataset_names, group_names=None):
        """
        Write `data` to HDF5 file, using `dataset_names` for datasets,
        and optionally `group_names` for groups.

        * data:          if `group_names` is None: np.ndarray of N data instances of size N x [...]
                         else list of np.ndarray of N data instances of size N x [...] each
        * dataset_names: list of strings
        * group_names:   None, or list of strings
        """
        with self._lock:
            assert isinstance(dataset_names, list), "`dataset_names` is of type {} and should be `list`".format(type(dataset_names))

            hf = self.hf
            if group_names is None:
                assert not isinstance(data, list),\
                       '`data` should be a `numpy.ndarray` when no groups are specified'
                self._write_datasets(hf, data, dataset_names)
            else:
                assert isinstance(data, list) and len(data) == len(group_names),\
                       'Each group name should correspond to a `data` list entry, `len(data)`={}, while `len(grou_names)`={}'.\
                       format(len(data), len(group_names))
                for i, name in enumerate(group_names):
                    group = hf.require_group(name)
                    self._write_datasets(group, data[i], dataset_names)

    def _read_datasets(self, reader, dataset_names):
        # internal to the class
        name = dataset_names[0]
        data0 = reader[name][...]
        data = np.empty((len(dataset_names),) + data0.shape,
                        dtype=data0.dtype)
        data[0,...] = data0
        for i in range(1, len(dataset_names)):
            if self.verbose:
                show_progress(i, len(data), prefix='Reading datasets')
            data[i, ...] = reader[dataset_names[i]][...]
        return data

    def read_data(self, dataset_names, group_names=None):
        """
        Read `dataset_names` from HDF5 file, optionally using `group_names`.

        * dataset_names: list of strings
        * group_names:   None, or list of strings
        :return:         np.ndarray
        """
        with self._lock:
            hf = self.hf
            assert isinstance(dataset_names, list), "`dataset_names` is of type {} and should be `list`".format(type(dataset_names))
            if group_names is None:
                return self._read_datasets(hf, dataset_names)
            else:
                return [self._read_datasets(hf[group_name], dataset_names)
                        for group_name in group_names]

    def read_data_random_group(self, dataset_names):
        """
        Reads `dataset_names` each one from a random group.
        At least one group must exist.

        * dataset_names: list of strings
        :return: np.ndarray
        """
        with self._lock:
            hf = self.hf
            group_names = np.array(self.group_names)
            idx = np.random.randint(0, len(group_names), len(dataset_names))
            names = ['{}/{}'.format(g, d) for g, d in zip(group_names[idx], dataset_names)]
            return self._read_datasets(hf, names)

    def summary(self, print_limit=100):
        """
        Prints a summary of the contents of the HDF5 file.
        Lists all groups and first `print_limit` datasets for each group.

        * print_limit: number of datasets to list per group.
        """
        size_bytes = os.path.getsize(self.file_name)
        print('File size: {:.3f} GB'.format(size_bytes/(2.0**30)))
        print('Groups:')

        hf = self.hf
        keys = list(hf.keys())
        for i, group_name in enumerate(keys):
            if i > print_limit:
                print('[...] showing %d of %d' % (print_limit, len(keys)))
                break
            print('*',group_name, '\b/')
            group = hf[group_name]
            try:
                group_keys = list(group.keys())
                print(' ', end=' ')
                for j, dataset_name in enumerate(group_keys):
                    if j > print_limit:
                        print('[...] showing %d of %d' % (print_limit, len(group_keys)))
                        break
                    print(dataset_name, end=' ')
                if len(group_keys):
                    print()
            except: pass

    @property
    def group_names(self):
        """
        :return: list of group names
        """
        return list(self.hf.keys())

    @property
    def dataset_names(self):
        """
        :return: list of dataset names
                 (if groups are present, from the first group)
        """
        values = list(self.hf.values())
        if isinstance(values[0], h5py._hl.dataset.Dataset):
            return list(self.hf.keys())
        else:
            return list(values[0].keys())

    # enable 'with' statements
    def __del__(self):
        self.__exit__(None, None, None)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.hf is not None:
            self.hf.flush()
            self.hf.close()
            del self.hf
            self.hf = None


def minmax(x):
    """
    Range of x.

    * x: list or np.ndarray
    :return: (min, max)
    """
    if not isinstance(x, np.ndarray):
        x = np.array(x)
    return x.min(), x.max()

def mapmm(x, new_range = (0, 1)):
    """
    Remap values in `x` to `new_range`.

    * x: np.ndarray
    * new_range: (min, max)
    :return: np.ndarray with values mapped to [new_range[0], new_range[1]]
    """
    mina, maxa = new_range
    if not type(x) == np.ndarray:
        x = np.asfarray(x, dtype='float32')
    if x.dtype != np.float32:
        x = x.astype(np.float32)
    minx, maxx = minmax(x)
    if minx < maxx:
        x = old_div((x-minx),(maxx-minx))*(maxa-mina)+mina
    return x

def plcc(x, y):
    """Pearson Linear Correlation Coefficient"""
    x, y = np.float32(x), np.float32(y)
    return scipy.stats.pearsonr(x, y)[0]

def srocc(xs, ys):
    """Spearman Rank Order Correlation Coefficient"""
    xranks = pd.Series(xs).rank()
    yranks = pd.Series(ys).rank()
    return plcc(xranks, yranks)
srcc = srocc

def dist2mos(x, scale=np.arange(1, 6)):
    """
    Find the MOS of a distribution of scores `x`, given a `scale`.
    e.g. x=[0,2,2], scale=[1,2,3] => MOS=2.5
    """
    x = old_div(x, np.reshape(np.sum(x*1., axis=1), (len(x), 1)))
    return np.sum(x * scale, axis=1)

def force_tuple(x):
    """Make tuple out of `x` if not already a tuple or `x` is None"""
    if x is not None and not isinstance(x, tuple):
        return (x,)
    else:
        return x

def force_list(x):
    """Make list out of `x` if not already a list or `x` is None"""
    if x is not None and not isinstance(x, list):
        return [x]
    else:
        return x

def make_dirs(filename):
    """
    Create directory structure described by `filename`.
    * filename: a valid system path
    """
    if not os.path.exists(os.path.dirname(filename)):
        try:
            os.makedirs(os.path.dirname(filename))
        except: pass

def updated_dict(d, only_existing=True, **updates):
    """
    Update dictionary `d` with `updates`, optionally changing `only_existing` keys.

    * d: dict
    * only_existing: do not add new keys, update existing ones only
    * updates: dict
    :return: updated dictionary
    """
    d = d.copy()
    if only_existing:
        common = {key: value for key, value in list(updates.items())
                  if key in list(d.keys())}
        d.update(common)
    else:
        d.update(updates)
    return d

def chunks(l, n):
    """Yields successive `n`-sized chunks from list `l`."""
    for i in range(0, len(l), n):
        yield l[i:i + n]

def pretty_print(d, indent=0, key_sep=':', trim=True):
    """
    Pretty print dictionary, recursively.

    * d: dict
    * indent: indentation amount
    * key_sep: separator printed between key and values
    * trim: remove redundant white space from printed values
    """
    if indent == 0 and not isinstance(d, dict):
        d = d.__dict__
    max_key_len = 0
    keys = list(d.keys())
    if isinstance(keys, list) and len(keys)>0:
        max_key_len = max([len(str(k)) for k in keys])
    for key, value in list(d.items()):
        equal_offset = ' '*(max_key_len - len(str(key)))
        print('\t' * indent + str(key) + key_sep, end=' ')
        if isinstance(value, dict):
            print()
            pretty(value, indent+1)
        else:
            value_str = str(value).strip()
            if trim:
                value_str = ' '.join(value_str.split())
            if len(value_str) > 70:
                value_str = value_str[:70] + ' [...]'
            print(equal_offset + value_str)

pretty = pretty_print

class ShortNameBuilder(Munch):
    """
    Utility for building short (file) names
    that contain multiple parameters.

    For internal use in ModelHelper.
    """
    def __init__(self, prefix='', sep=('', '_'),
                 max_len=32, **kwargs):
        self.__prefix  = prefix
        self.__sep     = sep
        self.__max_len = max_len
        super(ShortNameBuilder, self).__init__(**kwargs)

    def subset(self, selected_keys):
        subself = ShortNameBuilder(**self)
        for k in list(subself.keys()):
            if k not in selected_keys and \
               '_ShortNameBuilder' not in k:
                del subself[k]
        return subself

    def __call__(self, subset=None, **kwargs):
        self.update(**kwargs)
        return str(self)

    def __str__(self):
        def combine(k, v):
            k = str(k)[:self.__max_len]
            v = str(v)[:self.__max_len]
            return (k + self.__sep[0] + v if v else k)
        return self.__prefix + \
               self.__sep[1].join([combine(k, self[k])
                    for k in sorted(self.keys())
                    if '_ShortNameBuilder' not in k])

def check_keys_exist(new, old):
    """
    Check that keys in `new` dict existing in `old` dict.
    * new: dict
    * old: dict
    :return: exception if `new` keys don't existing in `old` ones

    Utility function used internally.
    """
    for key in list(new.keys()):
        if key not in list(old.keys()):
            raise Exception('Undefined parameter: "%s"' % key)

def get_sizes(x, array_marker='array'):
    """
    String representation of the shapes of arrays in lists/tuples.
    """
    if isinstance(x, list) or isinstance(x, tuple):
        content = ', '.join([get_sizes(_x_, array_marker) for _x_ in x])
        if isinstance(x, list):
            return '[' + content + ']'
        else:
            return '(' + content + ')'
    elif hasattr(x, 'shape'):
        return array_marker + '<' + ','.join(map(str, x.shape)) + '>'
    else: return '<1>'

def print_sizes(x, array_marker=''):
    """
    Prints get_sizes(x)
    """
    print(get_sizes(x, array_marker=array_marker))

def raw_confirm(message):
    """
    Ask for confirmation.

    * message: message to show
    :return: true if confirmation given, false otherwise
    """
    confirmation = input(message + " (y/[n])")
    if not confirmation:
        return False  # do not confirm by default
    else:
        return confirmation.lower()[0] == "y"

def update_config(obj, **kwargs):
    """
    Update configuration of Keras `obj` e.g. layer, model.

    * obj: object that has .get_config() and .from_config() methods.
    * kwargs: dict of updates
    :return: updated object
    """
    cfg = obj.get_config()
    cfg.update(**kwargs)
    return obj.from_config(cfg)

class Timer(object):
    def __init__(self, name=None):
        self.name = name

    def __enter__(self):
        self.tstart = time.time()

    def __exit__(self, type, value, traceback):
        if self.name:
            print(('[%s]' % self.name), end=' ')
        print('elapsed: %s seconds' % round(time.time() - self.tstart, 4))


def partition_rows(t, test_size=0.2, set_name='set',
                   set_values=['training', 'test'],
                   random_state=None, copy=True):
    if copy: t = t.copy()
    t = t.reset_index(drop=True)
    itrain, itest = train_test_split(list(range(len(t))),
                                     test_size=test_size,
                                     random_state=random_state)
    t.loc[itrain, set_name] = set_values[0]
    t.loc[itest,  set_name] = set_values[1]
    return t

def array_overlap(a, b):
    """
    Returns indices of overlapping values in two arrays.
    From: https://www.followthesheep.com/?p=1366
    """
    ia=np.argsort(a)
    ib=np.argsort(b)

    # use searchsorted:
    sort_left_a  = a[ia].searchsorted(b[ib], side='left')
    sort_right_a = a[ia].searchsorted(b[ib], side='right')
    #
    sort_left_b  = b[ib].searchsorted(a[ia], side='left')
    sort_right_b = b[ib].searchsorted(a[ia], side='right')

    # which values of b are also in a?
    inds_b = (sort_right_a-sort_left_a > 0).nonzero()[0]
    # which values of a are also in b?
    inds_a = (sort_right_b-sort_left_b > 0).nonzero()[0]

    return ia[inds_a], ib[inds_b]


# Print iterations progress
# Source: https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
def show_progress(iteration, total, prefix = '', suffix = '', decimals = 0,
                  length = 50, fill = '=', printEnd = "\r"):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    iteration_ = iteration+1
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration_ / float(total)))
    filledLength = int(length * iteration_ / total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end = printEnd)
    # Print New Line on Complete
    if iteration_ == total:
        print()

def download_archive(data_url, data_root, cache_root=None,
                     delete_archive=True, verbose=False, archive_type=None):
    archive_name = data_url.split('/')[-1]
    data_path = os.path.join(data_root, archive_name)
    make_dirs(data_root)

    if cache_root:
        make_dirs(cache_root)
        cache_path = os.path.join(cache_root, archive_name)
        if not os.path.exists(cache_path):
            if verbose:
                print('downloading {} via cache {}'.format(data_url, cache_root))
            urllib.request.urlretrieve(data_url, cache_path)
        if verbose:
            print('copying {} to {}'.format(archive_name, data_root))
        shutil.copy(cache_path, data_root)
    else:
        if not os.path.exists(data_path):
            if verbose:
                print('downloading {} to {}'.format(data_url, data_path))
            urllib.request.urlretrieve(data_url, data_path)

    if verbose:
        print('unpacking archive {} to {}'.format(archive_name, data_root))

    if  archive_type is None:
        archive_type = archive_name.split('.')[-1]

    if archive_type == 'zip':
        failed_command = subprocess.run(["unzip","-qq","-o",data_path,"-d",data_root], check=True)
    else:
        failed_command = subprocess.run(["tar","-xf",data_path,"-C",data_root], check=True)

    if failed_command.returncode:
        print("unpack failed: %d" % failed_command.returncode)

    if delete_archive:
        os.unlink(data_path)




from __future__ import print_function
from __future__ import division
from builtins import map
from past.utils import old_div
import os, sys
import tensorflow as tf
from tensorflow.keras import backend as K

# legacy, to be removed?
#if sys.version_info.major==3:
#    tf = K.tensorflow_backend.tf
#else:
#    tf = K.tf

# Keras configuration directives
def SetActiveGPU(number=0):
    """
    Set visibility of GPUs to the Tensorflow engine.

    * number: scalar or list of GPU indices
              e.g. 0 for the 1st GPU, or [0,2] for the 1st and 3rd GPU
    """
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    if not isinstance(number,list): number=[number]
    os.environ["CUDA_VISIBLE_DEVICES"] = ', '.join(map(str,number))
    print('Visible GPU(s):', os.environ["CUDA_VISIBLE_DEVICES"])

def GPUMemoryCap(fraction=1):
    """
    Limit the amount of GPU memory that can be used by an active kernel.

    * fraction: in [0, 1], 1 = the entire available GPU memory.
    """
    config = tf.ConfigProto()
    config.gpu_options.per_process_gpu_memory_fraction = fraction
    K.set_session(tf.Session(config=config))


# Metrics and losses

def plcc_tf(x, y):
    """PLCC metric"""
    xc = x - K.mean(x)
    yc = y - K.mean(y)
    return K.mean(xc*yc)/(K.std(x)*K.std(y) + K.epsilon())

def plcc_loss(x, y):
    """Loss version of `plcc_tf`"""
    return (1. - plcc_tf(x, y)) / 2.

def earth_mover_loss(y_true, y_pred):
    """
    Earth Mover's Distance loss.

    Reproduced from https://github.com/titu1994/neural-image-assessment/blob/master/train_inception_resnet.py
    """
    cdf_ytrue = K.cumsum(y_true, axis=-1)
    cdf_ypred = K.cumsum(y_pred, axis=-1)
    samplewise_emd = K.sqrt(K.mean(K.square(K.abs(cdf_ytrue - cdf_ypred)), axis=-1))
    return K.mean(samplewise_emd)

def make_loss(loss, **params_defa):
    def custom_loss(*args, **kwargs):
        kwargs.update(params_defa)
        return loss(*args, **kwargs)
    return custom_loss


def get_plcc_dist_tf(scores_array=[1., 2, 3, 4, 5]):
    """
    Function generator for `plcc_dist_tf`
    Computes the PLCC between the MOS values computed
    from pairs of distributions of scores. Used as a metric.

    * scores_array: scale values
    """
    def plcc_dist_tf(x, y):
        scores = K.constant(scores_array)
        xm = K.sum((x / K.reshape(K.sum(x, 1), [-1, 1])) * scores, 1)
        ym = K.sum((y / K.reshape(K.sum(y, 1), [-1, 1])) * scores, 1)
        x_sd = K.std(xm)
        y_sd = K.std(ym)
        xm_center = xm - K.mean(xm)
        ym_center = ym - K.mean(ym)
        return K.mean(xm_center*ym_center)/(x_sd*y_sd + 1e-3)
    return plcc_dist_tf

def get_plcc_dist_loss(scores_array=[1., 2, 3, 4, 5]):
    """Loss version of `plcc_dist_tf`"""
    plcc_dist_tf = get_plcc_dist_tf(scores_array)
    def plcc_dist_loss(x, y):
        return (1-plcc_dist_tf(x, y))/2
    return plcc_dist_loss




from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from builtins import zip, map, range
from past.utils import old_div
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt
from functools import reduce
from .model_helper import *


import sys
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.utils import plot_model
from tensorflow.keras.layers import *
from tensorflow.keras.applications.inception_v3 import InceptionV3
from tensorflow.keras.applications.densenet import DenseNet201
from tensorflow.keras.applications.resnet50 import ResNet50
from tensorflow.keras.applications.inception_resnet_v2 import InceptionResNetV2
from tensorflow.keras.applications.vgg16 import VGG16
from tensorflow.keras.applications.nasnet import NASNetMobile

source_module = {
                 InceptionV3:       keras.applications.inception_v3,
                 DenseNet201:       keras.applications.densenet,
                 ResNet50:          keras.applications.resnet50,
                 InceptionResNetV2: keras.applications.inception_resnet_v2,
                 VGG16:             keras.applications.vgg16,
                 NASNetMobile:      keras.applications.nasnet
                }

# correspondences between CNN model name and pre-processing function
process_input = {
                 InceptionV3:       keras.applications.inception_v3.preprocess_input,
                 DenseNet201:       keras.applications.densenet.preprocess_input,
                 ResNet50:          keras.applications.resnet50.preprocess_input,
                 InceptionResNetV2: keras.applications.inception_resnet_v2.preprocess_input,
                 VGG16:             keras.applications.vgg16.preprocess_input,
                 NASNetMobile:      keras.applications.nasnet.preprocess_input
                }

def fc_layers(input_layer,
              name               = 'pred',
              fc_sizes           = [2048, 1024, 256, 1],
              dropout_rates      = [0.25, 0.25, 0.5, 0],
              batch_norm         = False,
              l2_norm_inputs     = False,
              activation         = 'relu',
              initialization     = 'he_normal',
              out_activation     = 'linear',
              test_time_dropout  = False,
              **fc_params):
    """
    Add a standard fully-connected (fc) chain of layers (functional Keras interface)
    with dropouts on top of an input layer. Optionally batch normalize, add regularizers
    and an output activation.

    * input_layer: input layer to the chain
    * name: prefix to each layer in the chain
    * fc_sizes: list of number of neurons in each fc-layer
    * dropout_rates: list of dropout rates for each fc-layer
    * batch_norm: 0 (False) = no batch normalization (BN),
                  1 = do BN for all, 2 = do for all except the last, ...
    * l2_norm_inputs: normalize the `input_layer` with L2_norm
    * kernel_regularizer: optional regularizer for each fc-layer
    * out_activation: activation added to the last fc-layer
    :return: output layer of the chain
    """
    x = input_layer
    if l2_norm_inputs:
        x = Lambda(lambda x: tf.nn.l2_normalize(x, 1))(input_layer)

    assert dropout_rates is None or (len(fc_sizes) == len(dropout_rates)),\
           'Each FC layer should have a corresponding dropout rate'

    if activation.lower() == 'selu':
        dropout_call = AlphaDropout
    else:
        dropout_call = Dropout

    for i in range(len(fc_sizes)):
        if i < len(fc_sizes)-1:
            act = activation
            layer_type = 'fc%d' % i
        else:
            act  = out_activation
            layer_type = 'out'
        x = Dense(fc_sizes[i], activation=act,
                  name='%s_%s' % (name, layer_type),
                  kernel_initializer=initialization,
                  **fc_params)(x)
        if batch_norm > 0 and i < ( len(fc_sizes)-(batch_norm-1) ):
            x = BatchNormalization(name='%s_bn%d' % (name, i))(x)
        if dropout_rates is not None and dropout_rates[i] > 0:
            do_call = dropout_call(dropout_rates[i], name = '%s_do%d' % (name, i))
            if test_time_dropout:
                x = do_call(x, training=True)
            else:
                x = do_call(x)
    return x

def conv2d_bn(x, filters, num_row, num_col, padding='same',
              strides=(1, 1), name=None):
    """
    Utility function to apply conv + BN.

    * x:       input tensor.
    * filters: filters in `Conv2D`.
    * num_row: height of the convolution kernel.
    * num_col: width of the convolution kernel.
    * padding: padding mode in `Conv2D`.
    * strides: strides in `Conv2D`.
    * name:    name of the ops; will become `name + '_conv'`
               for the convolution and `name + '_bn'` for the
               batch norm layer.
    :return:   output tensor after applying `Conv2D` and `BatchNormalization`.

    Source: InceptionV3 Keras code
    """
    if name is not None:
        bn_name = name + '_bn'
        conv_name = name + '_conv'
    else:
        bn_name = None
        conv_name = None
    if K.image_data_format() == 'channels_first':
        bn_axis = 1
    else:
        bn_axis = 3
    x = Conv2D(filters, (num_row, num_col),
               strides=strides, padding=padding,
               use_bias=False, name=conv_name)(x)
    x = BatchNormalization(axis=bn_axis, scale=False, name=bn_name)(x)
    x = Activation('relu', name=name)(x)
    return x

def inception_block(x, size=768, name=''):
    channel_axis = 3

    branch1x1 = conv2d_bn(x, size, 1, 1, name=name+'branch_1x1')

    branch3x3 = conv2d_bn(x, size, 1, 1, name=name+'3x3_1x1')
    branch3x3_1 = conv2d_bn(branch3x3, old_div(size,2), 1, 3, name=name+'3x3_1x3')
    branch3x3_2 = conv2d_bn(branch3x3, old_div(size,2), 3, 1, name=name+'3x3_3x1')
    branch3x3 = concatenate(
        [branch3x3_1, branch3x3_2],
        axis=channel_axis,
        name=name+'branch_3x3')

    branch_pool = AveragePooling2D((3, 3), strides=(1, 1),
                                   padding='same',
                                   name=name+'avg_pool_2d')(x)
    branch_pool = conv2d_bn(branch_pool, size, 1, 1,
                            name=name+'branch_pool')

    y = concatenate(
        [branch1x1, branch3x3, branch_pool],
        axis=channel_axis,
        name=name+'mixed_final')
    return y

def model_inception_multigap(input_shape=(224, 224, 3), return_sizes=False,
                             indexes=list(range(11)), name = ''):
    """
    Build InceptionV3 multi-GAP model, that extracts narrow MLSP features.
    Relies on `get_inception_gaps`.

    * input_shape: shape of the input images
    * return_sizes: return the sizes of each layer: (model, gap_sizes)
    * indexes: indices to use from the usual GAPs
    * name: name of the model
    :return: model or (model, gap_sizes)
    """
    print('Loading InceptionV3 multi-gap with input_shape:', input_shape)

    model_base = InceptionV3(weights     = 'imagenet',
                             include_top = False,
                             input_shape = input_shape)
    print('Creating multi-GAP model')

    gap_name = name + '_' if name else ''

    feature_layers = [model_base.get_layer('mixed%d' % i)
                      for i in indexes]
    gaps = [GlobalAveragePooling2D(name=gap_name+"gap%d" % i)(l.output)
            for i, l in zip(indexes, feature_layers)]
    concat_gaps = Concatenate(name=gap_name+'concat_gaps')(gaps)

    model = Model(inputs  = model_base.input,
                  outputs = concat_gaps)
    if name:
        model.name = name

    if return_sizes:
        gap_sizes = [np.int32(g.get_shape()[1]) for g in gaps]
        return (model, gap_sizes)
    else:
        return model


def model_inceptionresnet_multigap(input_shape=(224, 224, 3),
                                   return_sizes=False):
    """
    Build InceptionResNetV2 multi-GAP model, that extracts narrow MLSP features.

    * input_shape: shape of the input images
    * return_sizes: return the sizes of each layer: (model, gap_sizes)
    :return: model or (model, gap_sizes)
    """
    print('Loading InceptionResNetV2 multi-gap with input_shape:', input_shape)

    model_base = InceptionResNetV2(weights='imagenet',
                                   include_top=False,
                                   input_shape=input_shape)
    print('Creating multi-GAP model')

    feature_layers = [l for l in model_base.layers if 'mixed' in l.name]
    gaps = [GlobalAveragePooling2D(name="gap%d" % i)(l.output)
            for i, l in enumerate(feature_layers)]
    concat_gaps = Concatenate(name='concatenated_gaps')(gaps)

    model = Model(inputs=model_base.input, outputs=concat_gaps)

    if return_sizes:
        gap_sizes = [np.int32(g.get_shape()[1]) for g in gaps]
        return (model, gap_sizes)
    else:
        return model

def model_inception_pooled(input_shape=(None, None, 3), indexes=list(range(11)),
                           pool_size=(5, 5), name='', return_sizes=False):
    """
    Returns the wide MLSP features, spatially pooled, from InceptionV3.
    Similar to `model_inception_multigap`.

    * input_shape: shape of the input images
    * indexes: indices to use from the usual pools
    * pool_size: spatial extend of the MLSP features
    * name: name of the model
    * return_sizes: return the sizes of each layer: (model, pool_sizes)
    :return: model or (model, pool_sizes)
    """
    print('Loading InceptionV3 multi-pooled with input_shape:', input_shape)
    model_base = InceptionV3(weights     = 'imagenet',
                             include_top = False,
                             input_shape = input_shape)
    print('Creating multi-pooled model')

    ImageResizer = Lambda(lambda x: tf.image.resize_area(x, pool_size),
                          name='feature_resizer')

    feature_layers = [model_base.get_layer('mixed%d' % i) for i in indexes]
    pools = [ImageResizer(l.output) for l in feature_layers]
    conc_pools = Concatenate(name='conc_pools', axis=3)(pools)

    model = Model(inputs  = model_base.input,
                  outputs = conc_pools)
    if name: model.name = name

    if return_sizes:
        pool_sizes = [[np.int32(x) for x in f.get_shape()[1:]] for f in pools]
        return model, pool_sizes
    else:
        return model

def model_inceptionresnet_pooled(input_shape=(None, None, 3), indexes=list(range(43)),
                                 pool_size=(5, 5), name='', return_sizes=False):
    """
    Returns the wide MLSP features, spatially pooled, from InceptionResNetV2.

    * input_shape: shape of the input images
    * indexes: indices of the modules to use
    * pool_size: spatial extend of the MLSP features
    * name: name of the model
    * return_sizes: return the sizes of each layer: (model, pool_sizes)
    :return: model or (model, pool_sizes)
    """

    print('Loading InceptionResNetV2 multi-pooled with input_shape:', input_shape)
    model_base = InceptionResNetV2(weights     = 'imagenet',
                                   include_top = False,
                                   input_shape = input_shape)
    print('Creating multi-pooled model')

    ImageResizer = Lambda(lambda x: tf.image.resize_area(x, pool_size),
                          name='feature_resizer')

    feature_layers = [l for l in model_base.layers if 'mixed' in l.name]
    feature_layers = [feature_layers[i] for i in indexes]
    pools = [ImageResizer(l.output) for l in feature_layers]
    conc_pools = Concatenate(name='conc_pools', axis=3)(pools)

    model = Model(inputs  = model_base.input, outputs = conc_pools)
    if name: model.name = name

    if return_sizes:
        pool_sizes = [[np.int32(x) for x in f.get_shape()[1:]] for f in pools]
        return model, pool_sizes
    else:
        return model


# ------------------
# RATING model utils
# ------------------

def test_rating_model(helper, ids_test=None,
                      output_layer=None, output_column=None,
                      groups=1, remodel=False, show_plot=True):
    """
    Test rating model performance. The output of the mode is assumed to be
    either a single score, or distribution of scores (can be a histogram).

    * helper:   ModelHelper object that contains the trained model
    * ids_test: optionally provide another set of data instances, replacing
                those in `helper.ids`
    * output_layer: the rating layer, if more than out output exists
                    if output_layer is None, we assume there is a single rating output
    * output_column: the column in ids that corresponds to the output
    * groups: if a number: repetitions of the testing procedure,
                           after which the results are averaged
              if list of strings: group names to repeat over,
                                  they are assumed to be different augmentations
    * remodel: change structure of the model when changing `output_layer`, or not
    * show_plot: plot results vs ground-truth
    :return: (y_true, y_pred, SRCC, PLCC)
    """
    print('Model outputs:', helper.model.output_names)
    if output_column is not None:
        print('Output column:', output_column)
    if ids_test is None: ids_test = helper.ids[helper.ids.set=='test']

    if isinstance(groups, numbers.Number):
        test_gen = helper.make_generator(ids_test,
                                         shuffle = False,
                                         fixed_batches = False)
        y_pred = helper.predict(test_gen, repeats=groups, remodel=remodel,
                                output_layer=output_layer)
        groups_list = list(range(groups))
    else:
        if isinstance(groups[0], (list, tuple, str)):
            groups_list = groups
        else:
            groups_list = list(map(str, groups))
        y_pred = []
        print('Predicting on groups:')
        for group in groups_list:
            print(group, end=' ')
            test_gen = helper.make_generator(ids_test, shuffle=False,
                                             fixed_batches = False,
                                             random_group  = False,
                                             group_names   = force_list(group))
            y_pred.append(helper.predict(test_gen, repeats=1, remodel=remodel,
                                         output_layer=output_layer))
        print()

    if isinstance(y_pred, list):
        y_pred = old_div(reduce(lambda x, y: (x+y), y_pred), len(y_pred))

    y_pred = np.squeeze(y_pred)

    if y_pred.ndim == 2:
        # for distributions
        print('Testing distributions')
        outputs = helper.gen_params.outputs
        y_pred = dist2mos(y_pred, scale=np.arange(1, len(outputs)+1))
        y_test = np.array(ids_test.loc[:, outputs])
        y_test = dist2mos(y_test, scale=np.arange(1, len(outputs)+1))
    else:
        # for single prediction
        print('Testing single prediction')
        if output_column is None:
            output = force_tuple(helper.gen_params.outputs)[0]
        else:
            output = output_column
        y_test = np.array(ids_test.loc[:,output])
        y_test = y_test.flatten()

    # in case the last batch was not used, and dataset size
    # is not a multiple of batch_size
    y_test = y_test[:len(y_pred)]

    SRCC_test = round(srocc(y_pred, y_test), 3)
    PLCC_test = round(plcc(y_pred, y_test), 3)
    print('SRCC/PLCC: {}/{}'.format(SRCC_test, PLCC_test))

    if show_plot:
        plt.plot(y_pred, y_test, '.', markersize=0.5)
        plt.xlabel('predicted'); plt.ylabel('ground-truth'); plt.show()
    return y_test, y_pred, SRCC_test, PLCC_test

def rating_metrics(y_true, y_pred, show_plot=True):
    """
    Print out performance measures given ground-truth (`y_true`) and predicted (`y_pred`) scalar arrays.
    """
    y_true, y_pred = np.array(y_true).squeeze(), np.array(y_pred).squeeze()
    p_plcc = np.round(plcc(y_true, y_pred),3)
    p_srcc = np.round(srcc(y_true, y_pred),3)
    p_mae  = np.round(np.mean(np.abs(y_true - y_pred)),3)
    p_rmse  = np.round(np.sqrt(np.mean((y_true - y_pred)**2)),3)

    if show_plot:
        print('SRCC: {} | PLCC: {} | MAE: {} | RMSE: {}'.\
              format(p_srcc, p_plcc, p_mae, p_rmse))
        plt.plot(y_true, y_pred,'.',markersize=1)
        plt.xlabel('ground-truth')
        plt.ylabel('predicted')
        plt.show()
    return (p_srcc, p_plcc, p_mae, p_rmse)

def get_train_test_sets(ids, stratify_on='MOS', test_size=(0.2, 0.2),
                        save_path=None, show_histograms=False,
                        stratify=False, random_state=None):
    """
    Devise a train/validation/test partition for a pd.DataFrame
    Adds a column 'set' to the input `ids` that identifies each row as one of:
    ['training', 'validation', 'test'] sets.

    The partition can be stratified based on a continuous variable,
    meaning that the variable is first quantized, and then a kind of
    'class-balancing' is performed based on the quantization.

    * ids: pd.DataFrame
    * stratify_on: column name from `ids` to stratify
                   ("class balance") the partitions on
    * test_size: ratio (or number) of rows to assign to each of
                 test and validation sets respectively
                 e.g. (<validation size>, <test size>) or
                      (<validation ratio>, <test ratio>)
    * save_path: optional save path for generated partitioned table
    * show_histograms: show histograms of the distribution of the
                       stratification column
    * stratify: do stratification
    * random_state: initialize random state with a fixed value,
                    for reproducibility
    :return: modified DataFrame
    """
    if not(isinstance(test_size, tuple) or
           isinstance(test_size, list)):
        test_size = (test_size, test_size)
    ids = ids.copy().reset_index(drop=True)
    idx = list(range(len(ids)))
    if not stratify:
        strata = None
    else:
        strata = np.int32(mapmm(ids.loc[:, stratify_on],
                                (0, stratify-1-1e-6)))

    idx_train_valid, idx_test = train_test_split(idx,
                                test_size=test_size[1],
                                random_state=random_state,
                                stratify=strata)
    strata_valid = None if strata is None else strata[idx_train_valid]
    idx_train, idx_valid = train_test_split(idx_train_valid,
                           test_size=test_size[0],
                           random_state=random_state,
                           stratify=strata_valid)

    print('Train size: {}, Validation size: {}, Test size: {}'.\
          format(len(idx_train), len(idx_valid), len(idx_test)))

    ids.loc[idx_train, 'set'] = 'training'
    ids.loc[idx_valid, 'set'] = 'validation'
    ids.loc[idx_test,  'set'] = 'test'
    if save_path is not None:
        ids.to_csv(save_path, index=False)

    if show_histograms:
        plt.hist(ids.loc[idx_train, stratify_on], density=True,
                 facecolor='g', alpha=0.75, bins=100)
        plt.show()
        plt.hist(ids.loc[idx_valid, stratify_on], density=True,
                 facecolor='b', alpha=0.75, bins=100)
        plt.show()
        plt.hist(ids.loc[idx_test, stratify_on], density=True,
                 facecolor='r', alpha=0.75, bins=100)
        plt.show()

    return ids


# ---------------
# To be OBSOLETED
# ---------------

def get_model_imagenet(net_name, input_shape=None,
                       plot=False, **kwargs):
    """Returns ImageNet models"""

    print('Loading model', net_name if isinstance(net_name, str)\
                                    else net_name.__name__)

    if net_name == ResNet50 or net_name == 'ResNet50':
        base_model = ResNet50(weights='imagenet', include_top=False,
                              input_shape=input_shape, **kwargs)
        feats = base_model.layers[-2]

    elif net_name == NASNetMobile or net_name == 'NASNetMobile':
        base_model = NASNetMobile(weights='imagenet',
                                  include_top=True,
                                  input_shape=input_shape, **kwargs)
        feats = base_model.layers[-3]

    elif net_name in list(source_module.keys()):
        base_model = net_name(weights='imagenet', include_top=False,
                              input_shape=input_shape, **kwargs)
        feats = base_model.layers[-1]

    else:
        raise Exception('Unknown model ' + net_name.__name__)

    gap = GlobalAveragePooling2D(name="final_gap")(feats.output)
    model = Model(inputs=base_model.input, outputs=gap)

    if plot: plot_model(base_model, show_shapes=True,
                        to_file='plots/{}_model.png'.format(net_name.__name__))
    return model, process_input[net_name]




from __future__ import print_function
from __future__ import absolute_import

# ignore warnings scikit-image, numpy > 1.17 in tensorflow
import warnings
warnings.filterwarnings('ignore',category=FutureWarning)

from . import generic, tensor_ops, image_utils, generators, model_helper, applications

# remove tensorflow warnings
import logging
class WarningFilter(logging.Filter):
    def filter(self, record):
        msg = record.getMessage()
        tf_warning1 = 'retry (from tensorflow.contrib.learn.python.learn.datasets.base)' in msg
        tf_warning2 = 'is deprecated' in msg
        return not (tf_warning1 or tf_warning2)
logger = logging.getLogger('tensorflow')
logger.addFilter(WarningFilter())

print('Loaded KU')




from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from builtins import str
from builtins import map
from builtins import zip
from builtins import range
from builtins import object
from past.utils import old_div
import math, os, numpy as np, glob
import scipy.ndimage.interpolation
import skimage.transform as transform
from numpy import interp
from numpy.random import rand
import PIL, random
from PIL import Image
import matplotlib.pyplot as plt
from .generic import *

class ImageAugmenter(object):
    """
    Provides methods to easily transform images.
    Meant for creating custom image augmentation functions for training Keras models.
    e.g. # randomly crop and flip left-right a 224x224 patch out of an image
         process_fn = lambda im: ImageAugmenter(im).crop((224,224)).fliplr().result
         # process_fn can be passed as an argument to keras.utils.Sequence objects (data generators)

    Provides various pre-defined customizable transformations, all randomizable:
    rotate, crop, fliplr, rescale, resize. The transformations can be easily chained.
    """
    def __init__(self, image, remap=False, verbose=False):
        """
        * image: image to be transformed, np.ndarray
        * remap: remap values to [0,1] for easier to apply transformations
                 these are mapped back to the initial range when .result is called
        * verbose: enable verbose prints
        """
        self._rotation_angle = 0
        self._original_range = minmax(image)
        self._remap = remap
        self.image = image if not self._remap else mapmm(image)
        self.verbose = verbose

    def rotate(self, angle, random=True):
        """
        Rotate self.image

        * angle: if `random` then rotation angle is a random value between [-`angle`, `angle`]
                 otherwise rotation angle is `angle`
        * random: random or by a fixed amount
        :return: self
        """
        if angle != 0 and random:
            # rotation angle is random between [-angle, angle]
            self._rotation_angle += (rand(1)-0.5)*2*angle
        else:
            self._rotation_angle += angle

        self.image = transform.rotate(self.image, self._rotation_angle,
                                      resize=False, cval=0,
                                      clip=True, preserve_range=True,
                                      mode='symmetric')
        return self

    def crop(self, crop_size, crop_pos=None, clip_rotation=False):
        """
        Crop a patch out of self.image. Relies on `extract_patch`.

        * crop_size: dimensions of the crop (pair of H x W)
        * crop_pos: if None, then a random crop is taken, otherwise the given `crop_pos` position is used
                    pair of relative coordinates: (0,0) = upper left corner, (1,1) = lower right corner
        * clip_rotation: clip a border around the image, such that the edge resulting from
                         having rotated the image is not visible
        :return: self
        """
        # equally crop in both dimensions if only one number is provided
        if not isinstance(crop_size, (list, tuple)):
            crop_size = [crop_size, crop_size]
        # if using a ratio crop, compute actual crop size
        crop_size = [np.int32(c*dim) if 0 < c <= (1+1e-6) else c\
                     for c, dim in zip(crop_size, self.image.shape[:2])]

        if self.verbose:
            print('image_size:', self.image.shape, 'crop_size:', crop_size)

        if crop_pos is None:
            if crop_size != self.image.shape[:2]:
                if clip_rotation:
                    lrr = largest_rotated_rect(self.image.shape[0],
                                               self.image.shape[1],
                                               math.radians(self._rotation_angle))
                    x, y = self.image.shape, lrr
                    border = (old_div((x[0]-y[0]),2), old_div((x[1]-y[1]),2))
                else:
                    border = (0, 0)
                self.image = extract_random_patch(self.image,
                                                  patch_size = crop_size,
                                                  border     = border)
        else:
            if crop_size != self.image.shape[:2]:
                self.image = extract_patch(self.image,
                                           patch_size     = crop_size,
                                           patch_position = crop_pos)
        return self

    def cropout(self, crop_size, crop_pos=None, fill_val=0):
        """
        Cropout a patch of self.image and replace it with `fill_val`. Relies on `cropout_patch`.

        * crop_size: dimensions of the cropout (pair of H x W)
        * crop_pos:  if None, then a random cropout is taken, otherwise the given `crop_pos` position is used
                     pair of relative coordinates: (0,0) = upper left corner, (1,1) = lower right corner
        * fill_val:  value to fill the cropout with
        :return:     self
        """
        # size of the cropout is equal in both dimensions if only one number is provided
        if not isinstance(crop_size, (list, tuple)):
            crop_size = [crop_size, crop_size]
        # if using a ratio cropout, compute actual cropout size
        crop_size = [np.int32(c*dim) if isinstance(c, float) and (0 < c) and (c <= 1.) else c\
                     for c, dim in zip(crop_size, self.image.shape[:2])]

        if self.verbose:
            print('image_size:', self.image.shape, 'crop_size:', crop_size, 'fill_val:', fill_val)

        if crop_pos is None:
            if crop_size != self.image.shape[:2]:
                border = (0, 0)
                self.image = cropout_random_patch(self.image,
                                                  patch_size = crop_size,
                                                  border     = border,
                                                  fill_val   = fill_val)
        else:
            if crop_size != self.image.shape[:2]:
                self.image = cropout_patch(self.image,
                                           patch_size     = crop_size,
                                           patch_position = crop_pos,
                                           fill_val       = fill_val)
        return self

    def fliplr(self, do=None):
        """
        Flip left-right self.image

        * do: if None, random flip, otherwise flip if do=True
        :return: self
        """
        if (do is None and rand(1) > 0.5) or do:
            self._rotation_angle = -self._rotation_angle
            self.image = np.fliplr(self.image)
        return self

    def rescale(self, target, proportion=1, min_dim=False):
        """
        Rescale self.image proportionally

        * target: (int) target resolution relative to the reference image resolution
                  taken to be either the height if `min_dim` else min(height, width)
                  (float) zoom level
        * proportion: modulating factor for the zoom
                      when proportion=1 target zoom is unchanged
                      when proportion=0 target zoom=1 (original size)
        * min_dim: bool
        :return: self
        """
        if isinstance(target, int): # target dimensions
            if not min_dim:
                # choose height for zoom
                zoom_target = self.image.shape[0]
            else:
                # choose minimum dimension
                zoom_target = min(self.image.shape[0],
                                  self.image.shape[1])
            zoom = old_div(1. * target, zoom_target)
        else:
            zoom = target
        zoom = (1-proportion) + proportion*zoom

        self.image = transform.rescale(self.image, zoom,
                                       preserve_range=True,
                                       mode='reflect')
        return self

    def resize(self, size, ensure_min=False, fit_frame=False):
        """
        Resize image to target dimensions, exact or fitting inside frame

        * size: (height, width) tuple
        * ensure_min: if true, `size` is the minimum size allowed
                      a dimension is not changed unless it is below the minimum size
        * fit_frame: size concerns the dimensions of the frame that the image is to be
                     fitted into, while preserving its aspect ratio
        :return: self
        """
        imsz = self.image.shape[:2] # (height, width)

        if not fit_frame:
            # resize if needed only
            if (not ensure_min and size != imsz) or\
               (ensure_min and (imsz[0] < size[0] or imsz[1] < size[1])):
                if ensure_min:
                    size = [max(a, b) for a, b in zip(imsz, size)]
                self.image = transform.resize(self.image, size,
                                              preserve_range=True)
        else:
            image_height, image_width = imsz
            frame_height, frame_width = size
            aspect_image = float(image_width)/image_height
            aspect_frame = float(frame_width)/frame_height
            if aspect_image > aspect_frame: # fit width
                target_width = frame_width
                target_height = frame_width / aspect_image
            else: # fit height
                target_height = frame_height
                target_width = frame_height * aspect_image

            target_width, target_height = int(np.round(target_width)), int(np.round(target_height))

            self.image = transform.resize(self.image, (target_height, target_width),
                                          preserve_range=True)
        return self

    @property
    def result_image(self):
        return array_to_img(self.result)

    @property
    def result(self):
        """
        :return: transformed image
        """
        if self._remap:
            return mapmm(self.image, self._original_range)
        else:
            return self.image


# -- utility functions --

def get_patch_dims(im, patch_size, patch_position):
    """
    Returns the dimensions of an image patch of size `patch_size`,
    with its center at `patch_position` expressed as a ratio of the image's H and W

    * im:             np.ndarray of size H x W x C
    * patch_size:     2-tuple of patch H x W
    * patch_position: 2-tuple containing patch location
                      (0,0) = upper left corner, (1,1) = lower right corner
    :return:          tuple of (upper left corner X coordinate,
                                upper left corner Y coordinate,
                                lower right corner X coordinate,
                                lower right corner Y coordinate)
    """
    Py, Px         = patch_position
    H, W, _        = im.shape
    H_crop, W_crop = patch_size

    H_crop, W_crop = min(H, H_crop), min(W, W_crop)
    Y_max, X_max   = (H - H_crop, W - W_crop)
    Yc, Xc         = H*Py, W*Px

    X0, Y0 = Xc-old_div(W_crop,2), Yc-old_div(H_crop,2)
    X0, Y0 = min(max(int(X0), 0), X_max),\
             min(max(int(Y0), 0), Y_max)

    return (X0, Y0, X0+W_crop, Y0+H_crop)

def extract_patch(im, patch_size=(224, 224),
                  patch_position=(0.5, 0.5)):
    """
    Extract a patch of size `patch_size`,
    with its center at `patch_position` expressed as a ratio of the image's H and W

    * im:             np.ndarray of size H x W x C
    * patch_size:     2-tuple of patch H x W
    * patch_position: 2-tuple containing patch location
                      (0,0) = upper left corner, (1,1) = lower right corner
    :return:          np.ndarray
    """
    (X0, Y0, X1, Y1) = get_patch_dims(im, patch_size, patch_position)
    return im[Y0:Y1, X0:X1, ]

def get_random_patch_dims(im, patch_size, border):
    """
    Returns the dimensions of a random image patch of size `patch_size`,
    with the center of the patch inside `border`

    * im:         np.ndarray of size H x W x C
    * patch_size: 2-tuple of patch H x W
    * border:     2-tuple of border H x W
    :return:      tuple of (upper left corner X coordinate,
                            upper left corner Y coordinate,
                            lower right corner X coordinate,
                            lower right corner Y coordinate)
    """
    H, W, _        = im.shape
    H_crop, W_crop = patch_size

    H_crop, W_crop = min(H, H_crop), min(W, W_crop)
    Y_min, X_min   = border
    Y_max, X_max   = (H - H_crop - Y_min, W - W_crop - X_min)

    if Y_max < Y_min:
        Y_min = old_div((H - H_crop), 2)
        Y_max = Y_min

    if X_max < X_min:
        X_min = old_div((W - W_crop), 2)
        X_max = X_min

    Y0 = int(np.round(rand(1)*(Y_max-Y_min) + Y_min))
    X0 = int(np.round(rand(1)*(X_max-X_min) + X_min))

    return (X0, Y0, X0+W_crop, Y0+H_crop)

def extract_random_patch(im, patch_size=(224, 224), border=(0, 0)):
    """
    Extract a random image patch of size `patch_size`,
    with the center of the patch inside `border`

    * im:         np.ndarray of size H x W x C
    * patch_size: 2-tuple of patch H x W
    * border:     2-tuple of border H x W
    :return:      np.ndarray
    """
    (X0, Y0, X1, Y1) = get_random_patch_dims(im, patch_size, border)
    return im[Y0:Y1, X0:X1, ]

def cropout_patch(im, patch_size=(224, 224),
                  patch_position=(0.5, 0.5), fill_val=0):
    """
    Cropout (replace) a patch of size `patch_size` with `fill_val`,
    with its center at `patch_position` expressed as a ratio of the image's H and W

    * im:             np.ndarray of size H x W x C
    * patch_size:     2-tuple of patch H x W
    * patch_position: 2-tuple containing patch location
                      (0,0) = upper left corner, (1,1) = lower right corner
    * fill_val:       value to fill into the cropout
    :return:          np.ndarray
    """
    (X0, Y0, X1, Y1) = get_patch_dims(im, patch_size, patch_position)
    im_ = im.copy()
    im_[Y0:Y1, X0:X1, ] = fill_val
    return im_

def cropout_random_patch(im, patch_size=(224, 224), border=(0, 0), fill_val=0):
    """
    Cropout (replace) a random patch of size `patch_size` with `fill_val`,
    with the center of the patch inside `border`

    * im:         np.ndarray of size H x W x C
    * patch_size: 2-tuple of patch H x W
    * border:     2-tuple of border H x W
    * fill_val:   value to fill into the cropout
    :return:      np.ndarray
    """
    (X0, Y0, X1, Y1) = get_random_patch_dims(im, patch_size, border)
    im_ = im.copy()
    im_[Y0:Y1, X0:X1, ] = fill_val
    return im_

# modified from stackoverflow
def largest_rotated_rect(w, h, angle):
    """
    Given a rectangle of size wxh that has been rotated by 'angle' (in
    radians), computes the width and height of the largest possible
    axis-aligned rectangle within the rotated rectangle.

    Original JS code by 'Andri' and Magnus Hoff from Stack Overflow
    Converted to Python by Aaron Snoswell
    """

    quadrant = int(math.floor(old_div(angle, (old_div(math.pi, 2))))) & 3
    sign_alpha = angle if ((quadrant & 1) == 0) else math.pi - angle
    alpha = (sign_alpha % math.pi + math.pi) % math.pi

    bb_w = w * math.cos(alpha) + h * math.sin(alpha)
    bb_h = w * math.sin(alpha) + h * math.cos(alpha)

    gamma = math.atan2(bb_w, bb_w) if (w < h) else math.atan2(bb_w, bb_w)

    delta = math.pi - alpha - gamma

    length = h if (w < h) else w

    d = length * math.cos(alpha)
    a = old_div(d * math.sin(alpha), math.sin(delta))

    y = a * math.cos(gamma)
    x = y * math.tan(gamma)

    return (
        bb_w - 2 * x,
        bb_h - 2 * y
    )

# shuffle tiled images

def image_to_tiles(im, num_patches):
    """
    Cut an image `im` into equal sized patches.

    * im: input image
    * num_patches: (num_vertical, num_horizontal)
    """
    H, W = im.shape[:2]
    H_count, W_count = num_patches
    patch_H, patch_W = int(H / H_count), int(W / W_count)
    tiles = [im[x:x+patch_H,y:y+patch_W] for x in range(0,H,patch_H)
                                         for y in range(0,W,patch_W)]
    return tiles

def tiles_to_image(tiles, num_patches):
    """
    Reconstruct an image from tiles, resulting from `image_to_tiles`.

    * tiles: list of image patches
    * num_patches: (num_vertical, num_horizontal)
    """
    H_count, W_count = num_patches
    rows = [np.concatenate(row_list, axis=1) for row_list in chunks(tiles, W_count)]
    return np.concatenate(rows)

def imshuffle(im, num_patches):
    """
    Cut image into patches, shuffle and return the shuffled reconstructed image.
    Uses `image_to_tiles` and `tiles_to_image`.

    * im: input image
    * num_patches: (num_vertical, num_horizontal)
    """
    t = image_to_tiles(im, num_patches)
    np.random.shuffle(t)
    return tiles_to_image(t, num_patches)

def imshuffle_pair(im1, im2, num_patches, ratio=0.5, flip=False):
    """
    Scramble patches coming from two images into a single image.
    Similar to `imshuffle`, but the patches come from images `im1` and `im2`.

    * im1, im2: input images, of equal size
    * num_patches: (num_vertical, num_horizontal) patches to divide each image in
    * ratio: fraction of patches to take from `im1`, the rest are taken from `im2`
    """
    t1 = image_to_tiles(im1, num_patches)
    t2 = image_to_tiles(im2, num_patches)
    np.random.shuffle(t1)
    np.random.shuffle(t2)
    counts = np.int32(np.round(len(t1)*ratio))
    t12 = t1[:counts] + t2[counts:]
    np.random.shuffle(t12)
    if flip:
        for i, _ in enumerate(t12):
            if rand(1) > 0.5:
                t12[i] = np.fliplr(t12[i])
    return tiles_to_image(t12, num_patches)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kuti']

package_data = \
{'': ['*']}

install_requires = \
['Pillow-PIL',
 'future',
 'h5py>=2.10.0,<3.0.0',
 'matplotlib',
 'munch',
 'numpy',
 'opencv-python',
 'pandas',
 'scikit-image>=0.16.2,<0.17.0',
 'scikit-learn>=0.22.1,<0.23.0',
 'scipy',
 'tensorflow-gpu>=1.14']

setup_kwargs = {
    'name': 'kuti',
    'version': '0.3.0',
    'description': 'Keras training management utilities.',
    'long_description': '# Keras utilities (Kuti)\n\nThe project contains utilities for image assessment development with Keras/Tensorflow, including utilities for model training, custom generators, image management and augmentation. This is a poetry package for [ku](https://github.com/subpic/ku).\n\n## Overview\n\nSome of the key components of each file:\n\n**`model_helper.py`**:\n\n* `ModelHelper`: Wrapper class that simplifies default usage of Keras for regression models.\n\n**`generators.py`**:\n\n* `DataGeneratorDisk`, `DataGeneratorHDF5`: Keras generators for on-disk images, and HDF5 stored features/images\n\n**`image_utils.py`**:\n\n* various utility functions for manipulating images (read, write to HDF5, batch resize, view batch)\n\n**`image_augmenter.py`**:\n\n* `ImageAugmenter`: Create custom image augmentation functions for training Keras models.\n\n**`generic.py`**:\n\n* `H5Helper`: Manage named data sets in HDF5 files, for us in Keras generators.\n* `ShortNameBuilder`: Utility for building short (file) names that contain multiple parameters.\n\n**`applications.py`**:\n\n* `model_inception_multigap`, `model_inceptionresnet_multigap`: Model definitions for extracting MLSP narrow features\n* `model_inception_pooled`, `model_inceptionresnet_pooled`: Model definitions for extracting MLSP wide features\n\nYou can find more information in the docstrings.\n',
    'author': 'Vlad Hosu',
    'author_email': 'subpic@gmail.com',
    'maintainer': 'Vlad Hosu',
    'maintainer_email': 'subpic@gmail.com',
    'url': 'https://github.com/subpic/ku/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from builtins import str
from builtins import map
from builtins import zip
from builtins import range
from builtins import object
from past.utils import old_div
import math, os, numpy as np, glob
import scipy.ndimage.interpolation
import skimage.transform as transform
from numpy import interp
from numpy.random import rand
import PIL
from PIL import Image
import matplotlib.pyplot as plt
from .generic import *
from .image_augmenter import ImageAugmenter

from tensorflow.keras.preprocessing.image import img_to_array, array_to_img, load_img

def imv(im, remap=True, figsize=None, **kwargs):
    """
    Display an image.
    """
    plt.figure(figsize=figsize)
    im_ = np.squeeze(im)
    if remap: im_ = mapmm(im_)
    if len(im_.shape)<3 or im_.shape[2]==1:
        kwargs.setdefault('cmap','gray')
    plt.imshow(im_, **kwargs);
    plt.axis('off');

def view_stack(ims, figsize=(20, 20), figshape=None,
               cmap='gray', vrange='all', **kwargs):
    """
    Display a stack or list of images using subplots.

    * ims: single np.ndarray of size [N x H x W x 3/1] or
           list of np.ndarray(s) of size [H x W x 3/1]
           (if list, np.stack is called first)
    * figsize: plt.figure(figsize=figsize)
    * figshape: (rows, cols) of the figure
                if None, the sizes are inferred
    * cmap:   color map, defaults to 'gray'
    * vrange: remap displayed value range:
              if 'all' set a global display range for the entire stack,
              if 'each' use a different display range for each image
    * kwargs: passed to `imshow` for each image
    """
    # get number of images
    if isinstance(ims, list): n = len(ims)
    else:                     n = ims.shape[0]

    if figshape is None:
        cols = int(np.ceil(np.sqrt(n)))
        rows = int(np.ceil(old_div(1.*n,cols)))
    else:
        rows, cols = figshape
    if vrange == 'all':
        if isinstance(ims, list):
            mm = list(map(minmax, ims))
            vrange = (min([p[0] for p in mm]),
                      max([p[1] for p in mm]))
        else:
            vrange = minmax(ims)
    elif vrange == 'each':
        vrange = (None, None)

    fig = plt.figure(figsize=figsize)
    for i in range(n):
        ax = fig.add_subplot(rows, cols, i+1)
        if isinstance(ims, list): im = ims[i]
        else:                     im = np.squeeze(ims[i, ...])
        ax.imshow(im, cmap=cmap,
                  vmin=vrange[0], vmax=vrange[1], **kwargs)
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

def read_image(image_path, image_size=1):
    """
    Read image from disk

    * image_path: full path to the image
    * image_size: resize image to specified size
                  can be a 2-tuple of (H, W) or a scalar zoom factor
    :return: np.ndarray
    """
    if type(image_size) == tuple:
        im = load_img(image_path, target_size=image_size)
        x = img_to_array(im)
    else:
        im = load_img(image_path)
        x = img_to_array(im)
        if not image_size == 1:
            new_size = list(map(int, (x.shape[0]*image_size, x.shape[1]*image_size)))
            x = transform.resize(x/255., new_size, mode='reflect')*255.
    return x

def write_image(im_arr, dst_path, fmt='jpg'):
    im = array_to_img(im_arr)
    make_dirs(dst_path)
    if fmt.lower() in ('jpg', 'jpeg'):
        im.save(dst_path, 'JPEG', quality=95)
    else:
        im.save(dst_path, fmt.upper())

def read_image_batch(image_paths, image_size=None, as_list=False):
    """
    Reads image array of np.uint8 and shape (num_images, *image_shape)

    * image_paths: list of image paths
    * image_size: if not None, image is resized
    * as_list: if True, return list of images,
               else return np.ndarray (default)
    :return: np.ndarray or list
    """
    images = None
    for i, image_path in enumerate(image_paths):
        im = load_img(image_path)
        if image_size is not None:
            im = im.resize(image_size, Image.LANCZOS)
        x = img_to_array(im).astype(np.uint8)
        if images is None:
            if not as_list:
                images = np.zeros((len(image_paths),) + x.shape,
                                  dtype=np.uint8)
            else: images = []
        if not as_list: images[i, ...] = x
        else: images.append(x)
    return images

def resize_image(x, size):
    """
    Resize image using skimage.transform.resize even when range is outside [-1,1].

    * x: np.ndarray
    * size: new size (H,W)
    :return: np.ndarray
    """
    if size != x.shape[:2]:
        minx, maxx = minmax(x)
        if maxx > 1 or minx < -1:
            x = mapmm(x)
        x = transform.resize(x, size, mode='reflect')
        if maxx > 1 or minx < -1:
            x = mapmm(x, (minx, maxx))
    return x

def glob_images(image_path, verbose=False, split=False, image_types =\
                ('*.jpg', '*.png', '*.bmp', '*.JPG', '*.BMP', '*.PNG')):

    # index all `image_types` in source path
    file_list = []
    for imtype in image_types:
        pattern = os.path.join(image_path, imtype)
        file_list.extend(glob.glob(pattern))

    if verbose:
        print('\nFound', len(file_list), 'images')

    if split:
        return list(zip(*[os.path.split(impath) for impath in file_list]))
    else:
        return file_list

def resize_folder(path_src, path_dst, image_size_dst=None,
                  overwrite=False, format_dst='jpg',
                  process_fn=None, jpeg_quality=95):
    """
    Resize an image folder, copying the resized images to a new destination folder.

    * path_src:       source folder path
    * path_dst:       destination folder path, created if does not exist
    * image_size_dst: destination image size
                      if None, convert the images to the new format only
    * overwrite:     enable to over-write destination images
    * format_dst:     format type, defaults to 'jpg'
    * process_fn:     apply custom processing function before resizing (optional) and saving
    * jpeg_quality:   quality level if saving as a JPEG image
    :return:          list of file names that triggered an error during read/resize/write
    """

    file_list = glob_images(path_src)
    make_dirs(path_dst)
    print('Resizing images from "{}" to "{}"'.format(path_src, path_dst))

    errors = []
    for i, file_path_src in enumerate(file_list):
        show_progress(i, len(file_list), prefix='Resizing')

        try:
            file_name = os.path.split(file_path_src)[1]
            (file_body, file_ext) = os.path.splitext(file_name)

            file_name_dst = file_body + '.' + format_dst.lower()
            file_path_dst = os.path.join(path_dst, file_name_dst)

            # check that image hasn't been already processed
            if overwrite or not os.path.isfile(file_path_dst):
                im = Image.open(file_path_src)
                if process_fn is not None:
                    im = process_fn(im)
                if image_size_dst is not None:
                    if isinstance(image_size_dst, float):
                        actual_size = [int(y*image_size_dst) for y in im.size]
                    else:
                        actual_size = image_size_dst
                    imx = im.resize(actual_size, Image.LANCZOS)
                else:
                    imx = im
                if format_dst.lower() in ('jpg', 'jpeg'):
                    imx.save(file_path_dst, 'JPEG', quality=jpeg_quality)
                else:
                    imx.save(file_path_dst, format_dst.upper())

        except Exception as e:
            print('Error saving', file_name)
            print('Exception:', e)
            errors.append(file_name)

    return errors

def check_images(image_dir, image_types =\
                 ('*.jpg', '*.png', '*.bmp', '*.JPG', '*.BMP', '*.PNG')):
    """
    Check which images from `image_dir` fail to read.

    * image_dir:   the image directory
    * image_types: match patterns for image file extensions, defaults:
                   ('*.jpg', '*.png', '*.bmp', '*.JPG', '*.BMP', '*.PNG')
    :return:       tuple of (list of failed image names, list of all image names)
    """
    file_list = glob_images(image_dir, image_types)
    image_names_err = []
    image_names_all = []
    for (i, file_path) in enumerate(file_list):
        show_progress(i, len(file_list), prefix='Checking')

        try:
            file_dir, file_name = os.path.split(file_path)
            file_body, file_ext = os.path.splitext(file_name)
            image_names_all.append(file_name)
            load_img(file_path) # try to load
        except:
            image_names_err.append(file_name)
    return (image_names_err, image_names_all)

def save_images_to_h5(image_path, h5_path, overwrite=False,
                      batch_size=32, image_size_dst=None):
    """
    Save a folder of JPEGs to an HDF5 file. Uses `read_image_batch` and `H5Helper`.

    * image_path: path to the source image folder
    * h5_path:    path to the destination HDF5 file; created if does not exist
    * overwrite: true/false
    * batch_size: number of images to read at a time
    * image_size_dst: new size of images, if not None
    """
    file_list = glob_images(image_path, ['*.jpg'])
#     file_list = glob.glob(os.path.join(image_path, '*.jpg'))
#     print('Found', len(file_list), 'JPG images')
    make_dirs(h5_path)
    print('Saving images from "{}" to "{}"'.format(image_path, h5_path))

    with H5Helper(h5_path, overwrite=overwrite) as h:
        for i, batch in enumerate(chunks(file_list, batch_size)):
            if i % 10 == 0:
                print(i*batch_size, end=' ')
            image_names = [str(os.path.basename(path)) for path in batch]
            images = read_image_batch(batch, image_size=image_size_dst)
            h.write_data(images, dataset_names=image_names)


def augment_folder(path_src, path_dst, process_gen, format_dst='jpg',
                   overwrite=False, verbose=True, simulate=False, file_list=None):
    """
    Augment an image folder, copying the augmented versions of the images to a new destination folder
    and returning a pd.DataFrame containing augmentation paths and parameters

    * path_src: source folder path
    * path_dst: destination folder path, created if does not exist
    * process_gen: generator returning `process_fn(im)` function to use
                   for each augmentation and a dictionary of arguments applied
    * format_dst: format type, defaults to 'jpg'
    * overwrite: enable to over-write destination images
    * verbose: display status information
    * simulate: generated ids pd.DataFrame, do not process/save images
    * file_list: list of files to be processed
    :return: ids type pd.DataFrame
    """

    if file_list is None:
        file_list = glob_images(path_src, verbose=verbose)

    if not simulate:
        make_dirs(path_dst)
    if verbose:
        print('Augmenting images from "{}" to "{}"'.format(path_src, path_dst))

    ids_list = []

    # args is a dict to be converted to a dir name, or is a string
    for process_fn, args in process_gen():
        if args:
            if isinstance(args, dict):
                args_str    = ', '.join(['{}:{}'.format(*a) for a in args.items()])
                args_values = list(args.values())
                args_keys   = list(args.keys())
            else:
                args_str = str(args)
                args_values = [args_str]
                args_keys   = ['folder']
            if verbose:
                print('Augmentation args "{}"'.format(args_str))
        else:
            args_str = ''

        for i, file_path_src in enumerate(file_list):
            if verbose:
                show_progress(i, len(file_list), prefix='Augmenting')


            file_name = os.path.split(file_path_src)[1]
            (file_body, file_ext) = os.path.splitext(file_name)
            file_name_dst = '{}/{}.{}'.format(args_str, file_body, format_dst.lower())
            file_path_dst = os.path.join(path_dst, file_name_dst)

            if not simulate:
                im = img_to_array(load_img(file_path_src))

                # check if image has already been processed
                if overwrite or not os.path.isfile(file_path_dst):
                    if isinstance(args, dict):
                        args_ = updated_dict(args, file_name=file_name,
                                            only_existing=False)
                        imx = array_to_img(process_fn(im, **args_))
                    else:
                        args_ = dict(file_name=file_name)
                        imx = array_to_img(process_fn(im, **args_))

                    make_dirs(file_path_dst)
                    if format_dst.lower() in ('jpg', 'jpeg'):
                        imx.save(file_path_dst, 'JPEG', quality=95)
                    else:
                        imx.save(file_path_dst, format_dst.upper())

            row_entry = [file_name, file_name_dst] + args_values
            ids_list.append(row_entry)


    ids_aug = pd.DataFrame(ids_list,
                           columns=['image_name','image_path']+args_keys)
    return ids_aug

